# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests builtin_fuzzers."""

import functools
import os
import unittest

from pyfakefs import fake_filesystem_unittest
import six

from clusterfuzz._internal.bot.fuzzers import builtin
from clusterfuzz._internal.bot.fuzzers import builtin_fuzzers
from clusterfuzz._internal.bot.tasks import fuzz_task
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class BuiltinFuzzersTest(unittest.TestCase):
  """builtin_fuzzers tests."""

  def test_get(self):
    """Tests get()."""
    self.assertIsInstance(
        builtin_fuzzers.get('libFuzzer'), builtin.BuiltinFuzzer)
    self.assertIsNone(builtin_fuzzers.get('does_not_exist'))


# pylint: disable=unused-argument
def _mock_fuzzer_run(output, num_generated, corpus_directory, self,
                     input_directory, output_directory, no_of_files):
  """Mock fuzzer run."""
  for i in range(num_generated):
    with open(os.path.join(output_directory, 'fuzz-%d' % i), 'w') as f:
      f.write('testcase')

  return builtin.BuiltinFuzzerResult(output, corpus_directory)


@test_utils.with_cloud_emulators('datastore')
class BuiltinFuzzersSetupTest(fake_filesystem_unittest.TestCase):
  """Test builtin fuzzers setup."""

  def setUp(self):
    """Setup for builtin fuzzer setup test."""
    helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir('/input')
    self.fs.create_dir('/output')
    self.fs.create_dir('/data-bundles')
    environment.set_value('DATA_BUNDLES_DIR', '/data-bundles')

    helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.libFuzzer.fuzzer.LibFuzzer.run',
        'clusterfuzz._internal.metrics.fuzzer_logs.get_bucket',
        'clusterfuzz._internal.google_cloud_utils.blobs.write_blob',
    ])

    self.fuzzer = data_types.Fuzzer(
        revision=1,
        file_size='builtin',
        source='builtin',
        name='libFuzzer',
        max_testcases=4,
        builtin=True)
    self.fuzzer.put()

    self.fuzzer_directory = os.path.join(
        environment.get_value('ROOT_DIR'), 'src', 'clusterfuzz', '_internal',
        'bot', 'fuzzers', 'libFuzzer')

    # Needed since local config is not available with fakefs.
    self.mock.get_bucket.return_value = None
    self.mock.write_blob.return_value = 'sample'

    environment.set_value('JOB_NAME', 'job')
    environment.set_value('INPUT_DIR', '/input')
    environment.set_value('MAX_TESTCASES', 4)

  def test_update_fuzzer(self):
    """Test fuzzer setup."""
    self.assertTrue(setup.update_fuzzer_and_data_bundles('libFuzzer'))
    self.assertEqual(self.fuzzer_directory, environment.get_value('FUZZER_DIR'))

  def test_generate_blackbox_fuzzers(self):
    """Test generate_blackbox_fuzzers (success)."""
    output = ('metadata::fuzzer_binary_name: fuzzer_binary_name\n')
    self.mock.run.side_effect = functools.partial(_mock_fuzzer_run, output, 4,
                                                  'corpus_dir')

    self.assertTrue(setup.update_fuzzer_and_data_bundles('libFuzzer'))

    session = fuzz_task.FuzzingSession('libFuzzer', 'job', 1)
    session.testcase_directory = '/output'
    session.data_directory = '/input'

    (error_occurred, testcase_file_paths, sync_corpus_directory,
     fuzzer_metadata) = session.generate_blackbox_testcases(
         self.fuzzer, self.fuzzer_directory, 4)
    self.assertEqual(1, len(self.mock.run.call_args_list))
    self.assertEqual(('/input', '/output', 4), self.mock.run.call_args[0][1:])

    self.assertFalse(error_occurred)
    six.assertCountEqual(self, [
        '/output/fuzz-0',
        '/output/fuzz-1',
        '/output/fuzz-2',
        '/output/fuzz-3',
    ], testcase_file_paths)

    self.assertEqual('corpus_dir', sync_corpus_directory)
    self.assertDictEqual({
        'fuzzer_binary_name': 'fuzzer_binary_name'
    }, fuzzer_metadata)

  def test_generate_blackbox_fuzzers_fail(self):
    """Test generate_blackbox_fuzzers (failure)."""
    self.mock.run.side_effect = builtin.BuiltinFuzzerException()
    self.assertTrue(setup.update_fuzzer_and_data_bundles('libFuzzer'))

    session = fuzz_task.FuzzingSession('libFuzzer', 'job', 1)
    session.testcase_directory = '/output'
    session.data_directory = '/input'

    with self.assertRaises(builtin.BuiltinFuzzerException):
      session.generate_blackbox_testcases(self.fuzzer, self.fuzzer_directory, 4)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""setup.py for libClusterFuzz."""
import setuptools

with open('README.md', 'r') as fh:
  long_description = fh.read()

setuptools.setup(
    name='clusterfuzz',
    version='2.5.2',
    author='ClusterFuzz authors',
    author_email='clusterfuzz-dev@googlegroups.com',
    description='ClusterFuzz',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/google/clusterfuzz',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        # TODO(ochang): Minimize dependencies.
        'google-api-python-client',
        'google-auth>=1.22.1',
        'google-auth-oauthlib',
        'google-cloud-core',
        'google-cloud-datastore==1.12.0',
        'google-cloud-logging',
        'google-cloud-monitoring',
        'google-cloud-ndb',
        'google-cloud-storage',
        'grpcio',
        'httplib2',
        'mozprocess',
        'oauth2client',
        'protobuf',
        'psutil',
        'PyYAML',
        'requests',
        'six',
    ],
    package_data={
        'clusterfuzz': ['lib-config/*', 'lib-config/**/*'],
    },
    python_requires='>=3.7',
    zip_safe=False,
)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trusted host."""

import sys
import threading
import time

from google.cloud import ndb
import grpc

from clusterfuzz._internal.base import untrusted
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.metrics import monitoring_metrics
from clusterfuzz._internal.protos import heartbeat_pb2
from clusterfuzz._internal.protos import heartbeat_pb2_grpc
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.protos import untrusted_runner_pb2_grpc
from clusterfuzz._internal.system import environment

from . import config

WAIT_TLS_CERT_SECONDS = 60
RPC_FAIL_WAIT_TIME = 10


class ChannelState(object):
  """The host's view of the channel state."""
  # Channel isn't ready for sending RPCs.
  NOT_READY = 0

  # Channel is ready for RPCS.
  READY = 1

  # The host found that the worker is in an inconsistent state. That is, we
  # detected that either the worker has a different source version, or if it
  # restarted without us knowing.
  INCONSISTENT = 2


class HostState(object):
  """The state of the host."""

  def __init__(self):
    self.channel = None
    self.stub = None
    self.heartbeat_thread = None
    self.expect_shutdown = False

    self.worker_start_time = None
    self.worker_bot_name = None

    # Protects access to |channel_state| and notifies when the state becomes
    # READY or INCONSISTENT.
    self.channel_condition = threading.Condition()
    self.channel_state = ChannelState.NOT_READY


_host_state = HostState()


class UntrustedRunnerStub(untrusted_runner_pb2_grpc.UntrustedRunnerStub):
  """Stub for making RPC calls.

  We override the generated stub because we need to wrap these RPC calls to add
  error handling/retry logic."""

  def __init__(self, channel):
    super(UntrustedRunnerStub, self).__init__(channel)

    # Don't wrap GetStatus() because it's used during connection state changes.
    # Don't wrap UpdateSource() because it can be expected to fail.

    # pylint: disable=invalid-name
    self.SetupRegularBuild = _wrap_call(self.SetupRegularBuild)
    self.RunProcess = _wrap_call(self.RunProcess)
    self.RunAndWait = _wrap_call(self.RunAndWait)
    self.CreateDirectory = _wrap_call(self.CreateDirectory)
    self.RemoveDirectory = _wrap_call(self.RemoveDirectory)
    self.ListFiles = _wrap_call(self.ListFiles)
    self.CopyFileTo = _wrap_call(self.CopyFileTo)
    self.CopyFileFrom = _wrap_call(self.CopyFileFrom)
    self.Stat = _wrap_call(self.Stat)
    self.UpdateEnvironment = _wrap_call(self.UpdateEnvironment)
    self.SymbolizeStacktrace = _wrap_call(self.SymbolizeStacktrace)
    self.GetFuzzTargets = _wrap_call(self.GetFuzzTargets)
    self.TerminateStaleApplicationInstances = _wrap_call(
        self.TerminateStaleApplicationInstances)

    # The following are RPCs that execute larger tasks. Don't retry these.
    self.PruneCorpus = _wrap_call(self.PruneCorpus, num_retries=0)
    self.ProcessTestcase = _wrap_call(self.ProcessTestcase, num_retries=0)
    self.EngineFuzz = _wrap_call(self.EngineFuzz, num_retries=0)
    self.EngineReproduce = _wrap_call(self.EngineReproduce, num_retries=0)
    # pylint: enable=invalid-name


def _check_channel_state(wait_time):
  """Check the channel's state."""
  with _host_state.channel_condition:
    if (_host_state.channel_state == ChannelState.READY or
        _host_state.channel_state == ChannelState.INCONSISTENT):
      # Nothing to do in these states.
      return _host_state.channel_state

    # The channel is not ready, so we wait for a (re)connect.
    _host_state.channel_condition.wait(wait_time)
    return _host_state.channel_state


def _wrap_call(func, num_retries=config.RPC_RETRY_ATTEMPTS):
  """Wrapper for stub calls to add error handling and retry logic."""

  def wrapped(*args, **kwargs):
    """Wrapper for adding retry logic."""
    for retry_attempt in range(num_retries + 1):
      # Wait for channel to (re)connect if necessary.
      state = _check_channel_state(config.RECONNECT_TIMEOUT_SECONDS)

      if state == ChannelState.INCONSISTENT:
        # No point retrying if the worker is inconsistent.
        monitoring_metrics.HOST_INCONSISTENT_COUNT.increment()
        logs.log_warn('Worker got into an inconsistent state.')
        host_exit_no_return(return_code=0)

      if state == ChannelState.NOT_READY:
        # Channel still isn't ready.
        logs.log_warn(
            'Channel failed to become ready within reconnect timeout.')
        if retry_attempt == num_retries:
          # Last attempt.
          host_exit_no_return()

        continue

      try:
        return func(*args, **kwargs)
      except grpc.RpcError as e:
        # For timeouts, which aren't fatal errors, resurface the right
        # exception.
        if 'TimeoutError' in repr(e):
          raise TimeoutError(str(e))

        if num_retries == 0:
          # Just re-raise the original exception if this RPC is not configured
          # for retries.
          raise

        logs.log_warn('Failed RPC: ' + repr(e))
        if retry_attempt == num_retries:
          # Last attempt.
          host_exit_no_return()

        time.sleep(RPC_FAIL_WAIT_TIME)

    return None

  return wrapped


def _do_heartbeat():
  """Heartbeat thread."""
  # grpc stubs and channels should be thread-safe.
  heartbeat_stub = heartbeat_pb2_grpc.HeartbeatStub(_host_state.channel)
  while True:
    try:
      heartbeat_stub.Beat(
          heartbeat_pb2.HeartbeatRequest(),
          timeout=config.HEARTBEAT_TIMEOUT_SECONDS)
    except grpc.RpcError as e:
      logs.log_warn('worker heartbeat failed: ' + repr(e))

    time.sleep(config.HEARTBEAT_INTERVAL_SECONDS)


def _get_host_worker_assignment():
  """Get the host worker assignment for the current host."""
  # This only needs to be called once before the host connects to the worker.
  # This is because the host->worker assignment algorithm should ensure that a
  # worker is reassigned only if it is also reimaged.
  #
  # If a worker is reimaged, then the host's connection state will be lost and
  # it will restart its run_bot.py instance to figure out which worker to
  # connect to again. We should never get into a case where worker re-assignment
  # happens without them being reimaged.
  key = ndb.Key(data_types.HostWorkerAssignment,
                environment.get_value('BOT_NAME'))
  return key.get()


def _get_root_cert(project_name):
  """Get the root TLS cert for connecting to the worker."""
  key = ndb.Key(data_types.WorkerTlsCert, project_name)
  tls_cert = key.get()
  if not tls_cert:
    return None

  assert tls_cert.cert_contents, 'Cert contents should not be empty.'
  return tls_cert.cert_contents


def _connect():
  """Initial connect to the worker."""
  worker_assignment = _get_host_worker_assignment()
  assert worker_assignment is not None
  assert worker_assignment.worker_name is not None
  assert worker_assignment.project_name is not None

  root_cert = _get_root_cert(worker_assignment.project_name)
  if not root_cert:
    logs.log_warn('TLS certs not yet generated.')
    time.sleep(WAIT_TLS_CERT_SECONDS)
    sys.exit(0)

  environment.set_value(
      'QUEUE_OVERRIDE',
      untrusted.platform_name(worker_assignment.project_name, 'linux'))

  server_name = worker_assignment.worker_name
  if not environment.get_value('LOCAL_DEVELOPMENT'):
    server_name += untrusted.internal_network_domain()

  _host_state.worker_bot_name = worker_assignment.worker_name

  credentials = grpc.ssl_channel_credentials(root_cert)
  _host_state.channel = grpc.secure_channel(
      '%s:%d' % (server_name, config.PORT),
      credentials=credentials,
      options=config.GRPC_OPTIONS)
  _host_state.stub = UntrustedRunnerStub(_host_state.channel)

  logs.log('Connecting to worker %s...' % server_name)
  _host_state.channel.subscribe(
      _channel_connectivity_changed, try_to_connect=True)

  channel_state = _check_channel_state(config.INITIAL_CONNECT_TIMEOUT_SECONDS)
  if channel_state == ChannelState.INCONSISTENT:
    logs.log_warn('Worker inconsistent on initial connect.')
    monitoring_metrics.HOST_INCONSISTENT_COUNT.increment()
    host_exit_no_return(return_code=0)

  if channel_state != ChannelState.READY:
    raise untrusted.HostException('Failed to connect to worker.')

  environment.set_value('WORKER_BOT_NAME', worker_assignment.worker_name)

  _host_state.heartbeat_thread = threading.Thread(target=_do_heartbeat)
  _host_state.heartbeat_thread.daemon = True
  _host_state.heartbeat_thread.start()


def _channel_connectivity_changed(connectivity):
  """Callback for channel connectivity changes."""
  try:
    with _host_state.channel_condition:
      if connectivity == grpc.ChannelConnectivity.READY:
        if _check_state():
          logs.log('Connected to worker.')
          _host_state.channel_state = ChannelState.READY
        else:
          _host_state.channel_state = ChannelState.INCONSISTENT

        _host_state.channel_condition.notify_all()
        return

      _host_state.channel_state = ChannelState.NOT_READY

    if connectivity == grpc.ChannelConnectivity.SHUTDOWN:
      if _host_state.expect_shutdown:
        # We requested a shutdown to update the source.
        logs.log('Worker shutting down.')
        return

      raise untrusted.HostException('Unrecoverable error.')
  except AttributeError:
    # Python sets all globals to None on shutdown. Ignore.
    logs.log('Shutting down.')
    return

  if connectivity == grpc.ChannelConnectivity.TRANSIENT_FAILURE:
    logs.log_warn('Transient failure detected on worker channel.')

  if connectivity == grpc.ChannelConnectivity.CONNECTING:
    logs.log('Reconnecting to worker.')


def _check_state():
  """Check that the worker's state is consistent with the host's knowledge."""
  try:
    status = stub().GetStatus(
        untrusted_runner_pb2.GetStatusRequest(),
        timeout=config.GET_STATUS_TIMEOUT_SECONDS)
  except grpc.RpcError:
    logs.log_error('GetStatus failed.')
    return False

  if status.revision != utils.current_source_version():
    logs.log_warn('Mismatching source revision: %s (host) vs %s (worker).' %
                  (utils.current_source_version(), status.revision))
    return False

  if _host_state.worker_bot_name != status.bot_name:
    logs.log_warn('Worker bot name invalid (IP changed?).')
    return False

  if _host_state.worker_start_time:
    if _host_state.worker_start_time == status.start_time:
      return True

    logs.log_warn('Worker start time changed.')
    return False

  _host_state.worker_start_time = status.start_time
  return True


def init():
  """Initialize channel to untrusted instance."""
  _connect()


def stub():
  """Return the UntrustedRunnerStub."""
  return _host_state.stub


def update_worker():
  """Update untrusted worker."""
  _host_state.expect_shutdown = True
  try:
    stub().UpdateSource(
        untrusted_runner_pb2.UpdateSourceRequest(),
        timeout=config.UPDATE_SOURCE_TIMEOUT_SECONDS)
  except grpc.RpcError:
    # Assume server got the shutdown request.
    pass


def host_exit_no_return(return_code=1):
  """Called when there is a host error."""
  if return_code:
    monitoring_metrics.HOST_ERROR_COUNT.increment({'return_code': return_code})

  # Always try to get the worker to exit too.
  update_worker()

  # Prevent exceptions during shutdown.
  _host_state.channel.unsubscribe(_channel_connectivity_changed)

  # This should bypass most exception handlers and avoid callers from catching
  # this incorrectly.
  logs.log('Shutting down host.', return_code=return_code)
  raise untrusted.HostException(return_code)


def is_initialized():
  """Return whether or not the host is initialized."""
  return _host_state.stub is not None



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for dictionary analysis and management."""

import os
import re

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import utils as fuzzer_utils
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

DICTIONARY_FILE_EXTENSION = '.dict'

# Name of the file in GCS containing recommended dictionary.
RECOMMENDED_DICTIONARY_FILENAME = (
    'recommended_dictionary%s' % DICTIONARY_FILE_EXTENSION)

# A comment string to separate recommended dictionary elements from manual ones.
RECOMMENDED_DICTIONARY_HEADER = '# Recommended dictionary stored in GCS.'

# Token to split a dictionary element analyzed and metadata.
TOKEN_ANALYZE_DICT_METADATA = ' # Score: '

# Tokens to detect "Recommended dictionary" section in the output.
TOKEN_RECOMMENDED_DICT_END = 'End of recommended dictionary.'
TOKEN_RECOMMENDED_DICT_START = 'Recommended dictionary.'

# Tokens to detect "useless dictionary" section in the output.
TOKEN_USELESS_DICT_END = 'End of useless dictionary elements.'
TOKEN_USELESS_DICT_START = 'Useless dictionary elements.'

DICTIONARY_PART_PATTERN = re.compile(r'([^"]+\s*=\s*)?(.*)')


def extract_dictionary_element(line):
  """Extract a dictionary element from the given string."""
  # An element should start and end with a double-quote.
  start_index = line.find('"')
  end_index = line.rfind('"')
  if start_index == -1 or end_index == -1 or start_index == end_index:
    return None

  element = line[start_index:end_index + 1]
  return element


def get_default_dictionary_path(fuzz_target_path):
  """Return default dictionary path."""
  return fuzzer_utils.get_supporting_file(fuzz_target_path,
                                          DICTIONARY_FILE_EXTENSION)


def get_dictionary_size(dictionary_content):
  """Calculate number of dictionary elements in the given string."""
  count = 0
  for line in dictionary_content.splitlines():
    if extract_dictionary_element(line):
      count += 1

  return count


def get_recommended_dictionary_gcs_path(fuzzer_name):
  """Generate a GCS url to a recommended dictionary.

  Returns:
    String representing GCS path for a dictionary.
  """
  bucket_name = environment.get_value('FUZZ_LOGS_BUCKET')
  bucket_subdirectory_name = 'dictionaries'
  recommended_dictionary_gcs_path = '/%s/%s/%s/%s' % (
      bucket_name, bucket_subdirectory_name, fuzzer_name,
      RECOMMENDED_DICTIONARY_FILENAME)

  return recommended_dictionary_gcs_path


def get_stats_for_dictionary_file(dictionary_path):
  """Calculate size of manual and recommended sections of given dictionary."""
  if not dictionary_path or not os.path.exists(dictionary_path):
    return 0, 0

  dictionary_content = utils.read_data_from_file(
      dictionary_path, eval_data=False).decode('utf-8')
  dictionaries = dictionary_content.split(RECOMMENDED_DICTIONARY_HEADER)

  # If there are any elements before RECOMMENDED_DICTIONARY_HEADER, those are
  # from "manual" dictionary stored in the repository.
  manual_dictionary_size = get_dictionary_size(dictionaries[0])
  if len(dictionaries) < 2:
    return manual_dictionary_size, 0

  # Any elements after RECOMMENDED_DICTIONARY_HEADER are recommended dictionary.
  recommended_dictionary_size = get_dictionary_size(dictionaries[1])
  return manual_dictionary_size, recommended_dictionary_size


def merge_dictionary_files(original_dictionary_path,
                           recommended_dictionary_path, merged_dictionary_path):
  """Merge a list of dictionaries with given paths into a singe dictionary."""
  if original_dictionary_path and os.path.exists(original_dictionary_path):
    merged_dictionary_data = utils.read_data_from_file(
        original_dictionary_path, eval_data=False).decode('utf-8')
  else:
    merged_dictionary_data = ''

  recommended_dictionary_lines = utils.read_data_from_file(
      recommended_dictionary_path,
      eval_data=False).decode('utf-8').splitlines()

  dictionary_lines_to_add = set()
  for line in recommended_dictionary_lines:
    if line not in merged_dictionary_data:
      dictionary_lines_to_add.add(line)

  merged_dictionary_data += '\n%s\n' % RECOMMENDED_DICTIONARY_HEADER

  merged_dictionary_data += '\n'.join(dictionary_lines_to_add)
  utils.write_data_to_file(merged_dictionary_data, merged_dictionary_path)


def _fix_dictionary_line(line, dict_path):
  """Correct a single dictionary line."""
  # Ignore blank and comment lines.
  if not line or line.strip().startswith('#'):
    return line

  match = DICTIONARY_PART_PATTERN.match(line)
  # We expect this pattern to match even invalid dictionary entries. Failures
  # to match should be treated as bugs in this function.
  if not match:
    raise errors.BadStateError(
        'Failed to correct dictionary line "{line}" in {path}.'.format(
            line=line, path=dict_path))

  name_part = match.group(1) or ''
  entry = match.group(2)

  # In some cases, we'll detect the user's intended entry as a token name. This
  # can happen if the user included unquoted tokens such as "!=" or ">=".
  if not entry and name_part:
    entry = name_part
    name_part = ''

  # Handle quote entries as a special case. This simplifies later logic.
  if entry == '"':
    entry = '"\\\""'

  if entry.startswith('"') and entry.endswith('"'):
    return name_part + entry

  # In this case, we know the entry is invalid. Escape any unescaped quotes
  # within it, then append quotes to the front and back.
  new_entry = ''
  prev_character = ''
  for character in entry:
    if character == '"' and prev_character != '\\':
      new_entry += '\\'
    new_entry += character
    prev_character = character

  new_entry = '"{entry}"'.format(entry=new_entry)
  return name_part + new_entry


def correct_if_needed(dict_path):
  """Corrects obvious errors such as missing quotes in a dictionary."""
  if not dict_path or not os.path.exists(dict_path):
    return

  content = utils.read_data_from_file(
      dict_path, eval_data=False).decode('utf-8')
  new_content = ''
  for current_line in content.splitlines():
    new_content += _fix_dictionary_line(current_line, dict_path) + '\n'

  # End of file newlines are inconsistent in dictionaries.
  if new_content.rstrip('\n') != content.rstrip('\n'):
    utils.write_data_to_file(new_content, dict_path)


class DictionaryManager(object):
  """Dictionary Manager."""

  def __init__(self, fuzzer_name):
    """Inits the DictionaryManager.

    Args:
      fuzzer_name: Name of the fuzzer.
    """
    self._fuzzer_name = fuzzer_name
    self._gcs_path = get_recommended_dictionary_gcs_path(self.fuzzer_name)

  @property
  def fuzzer_name(self):
    return self._fuzzer_name

  @property
  def gcs_path(self):
    return self._gcs_path

  def _compare_and_swap_gcs_dictionary(self, old_content, new_content):
    """Compare and swap implementation for dictionary stored in GCS. Of course,
    this function is not atomic, but window for race is acceptably small."""
    current_content = storage.read_data(self.gcs_path).decode('utf-8')
    if current_content != old_content:
      return False, current_content

    storage.write_data(new_content.encode('utf-8'), self.gcs_path)
    return True, old_content

  def download_recommended_dictionary_from_gcs(self, local_dict_path):
    """Download recommended dictionary from GCS to the given location.

    Args:
      local_dict_path: Path to a dictionary file on the disk.

    Returns:
      A boolean indicating whether downloading succeeded or not.
    """
    if environment.is_lib():
      return 0

    # When the fuzz target is initially created or when it has no new
    # coverage or dictionary recommendations, then we won't have a
    # recommended dictionary in GCS.
    if not storage.exists(self.gcs_path):
      return False

    if storage.copy_file_from(self.gcs_path, local_dict_path):
      return True

    logs.log('Downloading %s failed.' % self.gcs_path)
    return False

  def parse_recommended_dictionary_from_data(self, data):
    """Extract recommended dictionary entriees from the given string.

    Args:
      data: A string containing data from a fuzzer log.

    Returns:
      A set containing recommended dictionary lines.
    """
    log_lines = data.splitlines()
    return self.parse_recommended_dictionary_from_log_lines(log_lines)

  def parse_recommended_dictionary_from_log_lines(self, log_lines):
    """Extract recommended dictionary entriees from the given log lines.

    Args:
      log_lines: A list of strings from a fuzzer log.

    Returns:
      A set containing recommended dictionary lines.
    """
    # Process the lines in reverse order, as dictionary section is in the end.
    index = len(log_lines)
    while index:
      index -= 1
      if TOKEN_RECOMMENDED_DICT_END in log_lines[index]:
        # Found the section, now extract its entries.
        break

    recommended_entries = []
    while index:
      index -= 1
      if TOKEN_RECOMMENDED_DICT_START in log_lines[index]:
        # Beginning of the section reached, bail out.
        break

      element = extract_dictionary_element(log_lines[index])
      if element:
        recommended_entries.append(element)

    return recommended_entries

  def parse_useless_dictionary_from_data(self, data):
    """Extract useless dictionary entriees from the given string.

    Args:
      data: A string containing data from a fuzzer log.

    Returns:
      A set containing useless dictionary lines.
    """
    log_lines = data.splitlines()
    return self.parse_useless_dictionary_entries_from_log_lines(log_lines)

  def parse_useless_dictionary_entries_from_log_lines(self, log_lines):
    """Extract useless dictionary entries from the given log lines.

    Args:
      log_lines: A list of strings from a fuzzer log.

    Returns:
      A set containing useless dictionary lines.
    """
    # Process the lines in reverse order, as dictionary section is in the end.
    index = len(log_lines)
    while index:
      index -= 1
      if TOKEN_USELESS_DICT_END in log_lines[index]:
        # Found the section, now extract its entries.
        break

    useless_entries = []
    while index:
      index -= 1
      if TOKEN_USELESS_DICT_START in log_lines[index]:
        # Beginning of the section reached, bail out.
        break

      line = log_lines[index].split(TOKEN_ANALYZE_DICT_METADATA)[0]
      element = extract_dictionary_element(line)
      if element:
        useless_entries.append(element)

    return useless_entries

  def update_recommended_dictionary(self, new_dictionary):
    """Update recommended dictionary stored in GCS with new dictionary elements.

    Args:
      new_dictionary: A set of dictionary elements to be added into dictionary.

    Returns:
      A number of new elements actually added to the dictionary stored in GCS.
    """
    if environment.is_lib():
      return 0

    # If the dictionary does not already exist, then directly update it.
    if not storage.exists(self.gcs_path):
      storage.write_data('\n'.join(new_dictionary).encode('utf-8'),
                         self.gcs_path)
      return len(new_dictionary)

    # Read current version of the dictionary.
    old_dictionary_data = storage.read_data(self.gcs_path).decode('utf-8')

    # Use "Compare-and-swap"-like approach to avoid race conditions and also to
    # avoid having a separate job merging multiple recommended dictionaries.
    succeeded = False
    while not succeeded:
      # If old_dictionary_data is None, there is no dictionary in GCS yet, i.e.
      # it's empty. Otherwise, we parse it and use it.
      old_dictionary = set()
      if old_dictionary_data:
        old_dictionary = set(old_dictionary_data.splitlines())

      # Merge two dictionaries.
      new_dictionary |= old_dictionary
      if new_dictionary == old_dictionary:
        # "New dictionary" elements have been already added to GCS, bail out.
        return 0

      succeeded, old_dictionary_data = self._compare_and_swap_gcs_dictionary(
          old_dictionary_data, '\n'.join(new_dictionary))

    return len(new_dictionary) - len(old_dictionary)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions to apply build and revisions path overrides"""

import json
import os

from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.system import environment

PLATFORM_ID_URLS_FILENAME = 'config.json'
PLATFORM_ID_TO_BUILD_PATH_KEY = 'build_paths'
PLATFORM_ID_TO_REV_PATH_KEY = 'revisions_paths'

OVERRIDE_PATH_NOT_FOUND_ERROR = 'Could not find override path from config.'
OVERRIDE_CONFIG_NOT_JSON_ERROR = 'Could not read config as json.'
OVERRIDE_CONFIG_NOT_READ_ERROR = 'Could not import config file.'


def check_and_apply_overrides(curr_path, config_key):
  """Check if the given file points to a config, if so, use that to override
  any given paths"""
  if not curr_path:
    return curr_path
  if os.path.basename(curr_path) == PLATFORM_ID_URLS_FILENAME:
    curr_path = _apply_platform_id_overrides(environment.get_platform_id(),
                                             curr_path, config_key)
  return curr_path


def _apply_platform_id_overrides(platform_id, config_url, config_key):
  """read the `bucket_path`, parse as JSON, and map based on platform_id."""
  config_dict = _get_config_dict(config_url)
  path = _get_path_from_config(config_dict, config_key, platform_id)
  if not path:
    raise BuildOverrideError(OVERRIDE_PATH_NOT_FOUND_ERROR)
  return path


def _get_config_dict(url):
  """Read configs from a json and return them as a dict"""
  url_data = storage.read_data(url)
  if not url_data:
    raise BuildOverrideError(OVERRIDE_CONFIG_NOT_READ_ERROR)
  try:
    config_dict = json.loads(url_data)
  except ValueError:
    raise BuildOverrideError(OVERRIDE_CONFIG_NOT_JSON_ERROR)
  return config_dict


def _get_path_from_config(config_dict, config_key, platform_id):
  """Return True if a path override is present and return the override."""
  if config_key not in config_dict:
    return None
  return config_dict[config_key].get(platform_id)


class BuildOverrideError(Exception):
  """Build override error"""



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for libFuzzer script."""
# pylint: disable=unused-argument

import copy
import os
import shutil
import unittest

import pyfakefs.fake_filesystem_unittest as fake_fs_unittest

from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers import libfuzzer
from clusterfuzz._internal.bot.fuzzers import strategy_selection
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils

TESTDATA_PATH = os.path.join(os.path.dirname(__file__), 'libfuzzer_test_data')

BOT_NAME = 'test-bot'
BUILD_DIR = '/fake/build_dir'
FUZZ_INPUTS_DISK = '/fake/inputs-disk'
GSUTIL_PATH = '/fake/gsutil_path'
FAKE_ROOT_DIR = '/fake_root'

# An arbirtrary SHA1 sum.
ARBITRARY_SHA1_HASH = 'dd122581c8cd44d0227f9c305581ffcb4b6f1b46'


def _read_test_data(name):
  """Read test data."""
  data_path = os.path.join(TESTDATA_PATH, name)
  with open(data_path) as f:
    return f.read()


def read_data_from_file(file_path):
  """Reads data from file."""
  with open(file_path, 'rb') as file_handle:
    return file_handle.read().decode('utf-8')


def create_mock_popen(output,
                      corpus_path=None,
                      merge_corpus_path=None,
                      number_of_testcases=0,
                      return_code=0):
  """Creates a mock subprocess.Popen."""

  class MockPopen(object):
    """Mock subprocess.Popen."""
    commands = []
    testcases_written = []

    def __init__(self, command, *args, **kwargs):
      """Inits the MockPopen."""
      stdout = kwargs.pop('stdout', None)
      self.command = command
      self.commands.append(command)
      self.stdout = None
      self.return_code = return_code
      if hasattr(stdout, 'write'):
        self.stdout = stdout

    def _do_merge(self):
      """Mock merge."""
      if not corpus_path or not merge_corpus_path:
        return

      for filepath in os.listdir(corpus_path):
        shutil.copy(os.path.join(corpus_path, filepath), merge_corpus_path)

    def _write_fake_units(self):
      """Mock writing of new units."""
      for i in range(number_of_testcases):
        with open(os.path.join(corpus_path, str(i)), 'w') as f:
          f.write(str(i))

        self.testcases_written.append(str(i))

    def communicate(self, input_data=None):
      """Mock subprocess.Popen.communicate."""
      if '/fake/build_dir/fake_fuzzer' in self.command:
        if '-merge=1' in self.command:
          # Mock merge.
          self._do_merge()
        else:
          # Mock writing of new units.
          self._write_fake_units()

      if self.stdout:
        self.stdout.write(output)
      return None, None

    def poll(self, input_data=None):
      """Mock subprocess.Popen.poll."""
      return self.return_code

  return MockPopen


def mock_create_tmp_mount(base_dir):
  """Mock minijail._create_tmp_mount."""
  path = os.path.join(base_dir, 'TEMP')
  os.mkdir(path)
  return path


def mock_create_chroot_dir(base_dir):
  """Mock minijail._create_chroot_dir."""
  path = os.path.join(base_dir, 'CHROOT')
  os.mkdir(path)
  return path


def set_strategy_pool(strategies=None):
  """Helper method to create instances of strategy pools
  for patching use."""
  strategy_pool = strategy_selection.StrategyPool()

  if strategies is not None:
    for strategy_tuple in strategies:
      strategy_pool.add_strategy(strategy_tuple)
  return strategy_pool


class RecommendedDictionaryTest(fake_fs_unittest.TestCase):
  """Tests for dictionary processing."""

  def assert_compare_dictionaries(self, dict1, dict2):
    """Asserts that two given dictionaries contain the same elements."""
    # Order of elements in a merged dictionary is not guaranteed. Due to that,
    # we verify sizes of dictionaries and sets of their elements.
    self.assertEqual(len(dict1), len(dict2))
    self.assertEqual(set(dict1.splitlines()), set(dict2.splitlines()))

  def mock_download_recommended_dictionary_from_gcs(self, _,
                                                    local_dictionary_path):
    """Mock for DictionaryManager.download_recommended_dictionary_from_gcs."""
    engine_common.write_data_to_file(self.fake_gcs_dictionary_data,
                                     local_dictionary_path)
    return True

  def setUp(self):
    """Setup for recommended dictionary test."""
    # FIXME: Add support for Windows.
    if not environment.is_posix():
      self.skipTest('Process tests are only applicable for posix platforms.')

    self.data_directory = os.path.join(os.path.dirname(__file__), 'data')
    dictionaries_directory = os.path.join(self.data_directory, 'dictionaries')
    self.dictionary_from_repository_data = read_data_from_file(
        os.path.join(dictionaries_directory, 'dictionary_from_repository.dict'))
    self.expected_merged_dictionary_data = read_data_from_file(
        os.path.join(dictionaries_directory, 'expected_merged_dictionary.dict'))
    self.expected_gcs_only_merged_dictionary_data = read_data_from_file(
        os.path.join(dictionaries_directory,
                     'expected_gcs_only_merged_dictionary.dict'))
    self.fake_gcs_dictionary_data = read_data_from_file(
        os.path.join(dictionaries_directory, 'fake_gcs_dictionary.dict'))

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.dictionary_manager.DictionaryManager.'
        'download_recommended_dictionary_from_gcs',
        'os.getpid',
    ])
    self.mock.download_recommended_dictionary_from_gcs.side_effect = (
        self.mock_download_recommended_dictionary_from_gcs)
    self.mock.getpid.return_value = 1337

    test_utils.set_up_pyfakefs(self)
    self.work_directory_path = '/fake/fuzzers/'
    self.fuzz_inputs_path = '/fake/fuzzers/inputs'
    self.fuzz_inputs_disk_path = '/fake/fuzzers/inputs-disk'
    self.fs.create_dir(self.work_directory_path)
    self.fs.create_dir(self.fuzz_inputs_path)
    self.fs.create_dir(self.fuzz_inputs_disk_path)
    self.local_dictionaries_dir = os.path.join(self.work_directory_path,
                                               'dicts')
    self.fs.create_dir(self.local_dictionaries_dir)
    self.fuzzer_name = 'test_fuzzer'
    self.fuzzer_path = os.path.join(self.work_directory_path, self.fuzzer_name)

    test_helpers.patch_environ(self)
    environment.set_value('FAIL_RETRIES', '1')
    environment.set_value('FUZZ_INPUTS', self.fuzz_inputs_path)
    environment.set_value('FUZZ_INPUTS_DISK', self.fuzz_inputs_disk_path)

  def test_add_recommended_dictionary_no_merge(self):
    """Test dictionary processing when there is no local dictionary."""
    arguments = [
        '-max_len=80', '-rss_limit_mb=2048', '-timeout=25',
        '-artifact_prefix=/fake/', '-max_total_time=2950',
        '-print_final_stats=1', '/fake/inputs-disk/temp-1337/new',
        '/fake/corpus_basic'
    ]

    libfuzzer.add_recommended_dictionary(arguments, self.fuzzer_name,
                                         self.fuzzer_path)

    expected_dictionary_path = '%s.dict.merged' % self.fuzzer_path

    # Check '-dict' argument that should be added by
    # add_recommended_dictionary().
    expected_dictionary_argument = '-dict=%s' % expected_dictionary_path
    self.assertTrue(expected_dictionary_argument in arguments)

    # Check the dictionary contents.
    dictionary_data = read_data_from_file(expected_dictionary_path)
    self.assert_compare_dictionaries(
        dictionary_data, self.expected_gcs_only_merged_dictionary_data)

  def test_add_recommended_dictionary_with_merge(self):
    """Test dictionary processing when there is a local dictionary."""
    dictionary_path = os.path.join(self.work_directory_path,
                                   'dictionary_from_repository.dict')
    self.fs.create_file(
        dictionary_path, contents=self.dictionary_from_repository_data)

    dictionary_argument = '-dict=%s' % dictionary_path
    expected_arguments = [
        '-max_len=80', '-rss_limit_mb=2048', '-timeout=25', dictionary_argument,
        '-artifact_prefix=/fake/', '-max_total_time=2950',
        '-print_final_stats=1', '/fake/inputs-disk/temp-1337/new',
        '/fake/corpus_basic'
    ]

    actual_arguments = copy.deepcopy(expected_arguments)

    # The function call below is expected to modify actual_arguments list.
    libfuzzer.add_recommended_dictionary(actual_arguments, self.fuzzer_name,
                                         self.fuzzer_path)

    # The dictionary argument is expected to be removed and added to the end.
    expected_arguments.remove(dictionary_argument)
    expected_arguments.append(dictionary_argument + '.merged')
    self.assertEqual(actual_arguments, expected_arguments)

    # The dictionary should content merged data.
    updated_dictionary_data = read_data_from_file(dictionary_path + '.merged')
    self.assert_compare_dictionaries(updated_dictionary_data,
                                     self.expected_merged_dictionary_data)

  def test_download_recommended_dictionary_with_merge(self):
    """Test downloading of recommended dictionary."""
    arguments = [
        '-max_len=80', '-rss_limit_mb=2048', '-timeout=25',
        '-artifact_prefix=/fake/', '-max_total_time=2950',
        '-print_final_stats=1', '/fake/inputs-disk/temp-1337/new',
        '/fake/corpus_basic'
    ]

    libfuzzer.add_recommended_dictionary(arguments, self.fuzzer_name,
                                         self.fuzzer_path)
    self.assertIn(
        '/fake/fuzzers/inputs-disk/temp-1337/recommended_dictionary.dict',
        self.mock.download_recommended_dictionary_from_gcs.call_args[0])


class IsSha1HashTest(unittest.TestCase):
  """Tests for is_sha1_hash."""

  def test_non_hashes(self):
    """Tests that False is returned for non hashes."""
    self.assertFalse(libfuzzer.is_sha1_hash(''))
    self.assertFalse(libfuzzer.is_sha1_hash('z' * 40))
    self.assertFalse(libfuzzer.is_sha1_hash('a' * 50))
    fake_hash = str('z' + ARBITRARY_SHA1_HASH[1:])
    self.assertFalse(libfuzzer.is_sha1_hash(fake_hash))

  def test_hash(self):
    """Tests that False is returned for a real hash."""
    self.assertTrue(libfuzzer.is_sha1_hash(ARBITRARY_SHA1_HASH))


class MoveMergeableUnitsTest(fake_fs_unittest.TestCase):
  """Tests for move_mergeable_units."""
  CORPUS_DIRECTORY = '/corpus'
  MERGE_DIRECTORY = '/corpus-merge'

  def setUp(self):
    test_utils.set_up_pyfakefs(self)

  def move_mergeable_units(self):
    """Helper function for move_mergeable_units."""
    libfuzzer.move_mergeable_units(self.MERGE_DIRECTORY, self.CORPUS_DIRECTORY)

  def test_duplicate_not_moved(self):
    """Tests that a duplicated file is not moved into the corpus directory."""
    self.fs.create_file(
        os.path.join(self.CORPUS_DIRECTORY, ARBITRARY_SHA1_HASH))
    merge_corpus_file = os.path.join(self.MERGE_DIRECTORY, ARBITRARY_SHA1_HASH)
    self.fs.create_file(merge_corpus_file)
    self.move_mergeable_units()
    # File will be deleted from merge directory if it isn't a duplicate.
    self.assertTrue(os.path.exists(merge_corpus_file))

  def test_new_file_moved(self):
    """Tests that a new file is moved into the corpus directory."""
    # Make a file that looks like a sha1 hash but is different from
    # ARBITRARY_SHA1_HASH.
    filename = ARBITRARY_SHA1_HASH.replace('d', 'a')
    self.fs.create_file(os.path.join(self.CORPUS_DIRECTORY, filename))
    # Create an arbitrary file with a hash name that is different from this
    # filename.
    merge_corpus_file = os.path.join(self.MERGE_DIRECTORY, ARBITRARY_SHA1_HASH)
    self.fs.create_file(merge_corpus_file)
    self.move_mergeable_units()
    # File will be deleted from merge directory if it isn't a duplicate.
    self.assertFalse(os.path.exists(merge_corpus_file))
    self.assertTrue(
        os.path.exists(os.path.join(self.CORPUS_DIRECTORY, filename)))


class SelectGeneratorTest(unittest.TestCase):
  """Tests for _select_generator."""
  FUZZER_PATH = '/fake/fuzzer_path'

  def setUp(self):
    self.pool = strategy_selection.generate_default_strategy_pool(
        strategy_list=strategy.LIBFUZZER_STRATEGY_LIST, use_generator=True)
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.is_lpm_fuzz_target',
        'clusterfuzz._internal.bot.fuzzers.strategy_selection.StrategyPool.do_strategy'
    ])
    self.mock.do_strategy.return_value = True
    self.mock.is_lpm_fuzz_target.return_value = True

  def test_lpm_fuzz_target(self):
    self.assertEqual(engine_common.Generator.NONE,
                     engine_common.select_generator(self.pool,
                                                    self.FUZZER_PATH))  # pylint: disable=protected-access


if __name__ == '__main__':
  unittest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the issue_tracker module."""

import unittest

import mock
import six

from libs.issue_management.issue_tracker import IssueTracker
from libs.issue_management.issue_tracker import LabelStore


class LabelStoreTest(unittest.TestCase):
  """LabelStore tests."""

  def test_init_and_iter(self):
    """Test initializing and iterating LabelStore."""
    store = LabelStore(['label1', 'label2'])
    six.assertCountEqual(self, [], store.added)
    six.assertCountEqual(self, [], store.removed)
    six.assertCountEqual(self, ['label1', 'label2'], list(store))

  def test_add(self):
    """Test adding items."""
    store = LabelStore()
    store.add('laBel1')
    six.assertCountEqual(self, ['laBel1'], store)
    six.assertCountEqual(self, ['laBel1'], store.added)
    six.assertCountEqual(self, [], store.removed)

    store.add('label2')
    six.assertCountEqual(self, ['laBel1', 'label2'], store)
    six.assertCountEqual(self, ['laBel1', 'label2'], store.added)
    six.assertCountEqual(self, [], store.removed)

    store.add('labEl2')
    six.assertCountEqual(self, ['laBel1', 'labEl2'], store)
    six.assertCountEqual(self, ['laBel1', 'labEl2'], store.added)
    six.assertCountEqual(self, [], store.removed)

  def test_remove(self):
    """Test removing items."""
    store = LabelStore(['laBel1', 'label2', 'Label3'])
    store.remove('Label1')
    six.assertCountEqual(self, ['label2', 'Label3'], store)
    six.assertCountEqual(self, [], store.added)
    six.assertCountEqual(self, ['Label1'], store.removed)

    store.remove('Label2')
    six.assertCountEqual(self, ['Label3'], store)
    six.assertCountEqual(self, [], store.added)
    six.assertCountEqual(self, ['Label1', 'Label2'], store.removed)

    store.remove('LaBel2')
    six.assertCountEqual(self, ['Label3'], store)
    six.assertCountEqual(self, [], store.added)
    six.assertCountEqual(self, ['Label1', 'Label2'], store.removed)

    store.remove('Label4')
    six.assertCountEqual(self, ['Label3'], store)
    six.assertCountEqual(self, [], store.added)
    six.assertCountEqual(self, ['Label1', 'Label2'], store.removed)

  def test_add_and_remove(self):
    """Test both adding and removing."""
    store = LabelStore(['laBel1', 'label2', 'Label3'])
    store.remove('Label1')
    six.assertCountEqual(self, ['label2', 'Label3'], store)
    six.assertCountEqual(self, [], store.added)
    six.assertCountEqual(self, ['Label1'], store.removed)

    store.add('label1')
    six.assertCountEqual(self, ['label1', 'label2', 'Label3'], store)
    six.assertCountEqual(self, [], store.added)
    six.assertCountEqual(self, [], store.removed)

    store.remove('Label1')
    store.add('label4')
    six.assertCountEqual(self, ['label2', 'Label3', 'label4'], store)
    six.assertCountEqual(self, ['label4'], store.added)
    six.assertCountEqual(self, ['Label1'], store.removed)

  def test_reset(self):
    """Test reset."""
    store = LabelStore(['laBel1', 'label2', 'Label3'])
    store.add('label4')
    store.add('label5')
    store.remove('label1')

    store.reset_tracking()
    six.assertCountEqual(self, [], store.added)
    six.assertCountEqual(self, [], store.removed)

  def test_remove_by_prefix(self):
    """Test remove_by_prefix."""
    store = LabelStore(['p-0', 'P-1', 'Q-2'])
    store.remove_by_prefix('p-')
    six.assertCountEqual(self, ['Q-2'], store)

  def test_in(self):
    """Test in operator."""
    store = LabelStore(['laBel1', 'label2', 'Label3'])
    self.assertTrue('label1' in store)
    self.assertTrue('laBel2' in store)
    self.assertTrue('labeL3' in store)
    self.assertFalse('label' in store)


class TestIssueTracker(IssueTracker):
  """Test issue tracker."""

  def __init__(self):
    self.issues = {}

  def get_issue(self, issue_id):
    return self.issues.get(str(issue_id))


class GetOriginalIssueTest(unittest.TestCase):
  """Tests for get_original_issue."""

  def setUp(self):
    self.issue_tracker = TestIssueTracker()
    self.issue_tracker.issues = {
        '1': mock.Mock(id=1, merged_into=2),
        '2': mock.Mock(id=2, merged_into=3),
        '3': mock.Mock(id=3, merged_into=None),
        '4': mock.Mock(id=4, merged_into=4),
        '5': mock.Mock(id=5, merged_into=6),
        '6': mock.Mock(id=6, merged_into=5),
    }

  def test_basic(self):
    """Basic tests."""
    self.assertEqual(3, self.issue_tracker.get_original_issue(1).id)
    self.assertEqual(3, self.issue_tracker.get_original_issue(2).id)
    self.assertEqual(3, self.issue_tracker.get_original_issue(3).id)

  def test_circular_merge(self):
    """Test circular merge."""
    self.assertEqual(4, self.issue_tracker.get_original_issue(4).id)
    self.assertEqual(6, self.issue_tracker.get_original_issue(5).id)
    self.assertEqual(5, self.issue_tracker.get_original_issue(6).id)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generate new inputs using ML RNN model."""
# pylint: disable=g-statement-before-imports
try:
  # ClusterFuzz dependencies.
  from clusterfuzz._internal.base import modules
  modules.fix_module_search_paths()
except ImportError:
  pass

import os
import sys

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers.ml.rnn import constants
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import shell

# Model script directory absolute path.
ML_RNN_SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))

# Maximum number of new units to generate.
GENERATION_MAX_COUNT = 5000


def download_model_from_gcs(local_model_directory, fuzzer_name):
  """Pull model from GCS bucket and put them in specified model directory."""
  # ML model is stored in corpus bucket.
  gcs_corpus_bucket = environment.get_value('CORPUS_BUCKET')
  if not gcs_corpus_bucket:
    logs.log('Corpus bucket is not set. Skip generation.')
    return False

  # Get cloud storage path.
  # e.g. gs://clusterfuzz-corpus/rnn/libpng_read_fuzzer
  gcs_model_directory = 'gs://%s/%s/%s' % (
      gcs_corpus_bucket, constants.RNN_MODEL_NAME, fuzzer_name)

  logs.log('GCS model directory for fuzzer %s is %s.' % (fuzzer_name,
                                                         gcs_model_directory))

  # RNN model consists of two files.
  data_filename = constants.RNN_MODEL_NAME + constants.MODEL_DATA_SUFFIX
  index_filename = constants.RNN_MODEL_NAME + constants.MODEL_INDEX_SUFFIX

  # Cloud file paths.
  gcs_data_path = '%s/%s' % (gcs_model_directory, data_filename)
  gcs_index_path = '%s/%s' % (gcs_model_directory, index_filename)

  # Check if model exists.
  if not (storage.exists(gcs_data_path) and storage.exists(gcs_index_path)):
    logs.log('ML RNN model for fuzzer %s does not exist. Skip generation.' %
             fuzzer_name)
    return False

  # Local file paths.
  local_data_path = os.path.join(local_model_directory, data_filename)
  local_index_path = os.path.join(local_model_directory, index_filename)

  # Download model files.
  result = (
      storage.copy_file_from(gcs_data_path, local_data_path) and
      storage.copy_file_from(gcs_index_path, local_index_path))

  if not result:
    logs.log('Failed to download RNN model for fuzzer %s. Skip generation.' %
             fuzzer_name)
    return False

  return True


def prepare_model_directory(fuzzer_name):
  """Prepare model directory, and return model path.

  Args:
    fuzzer_name: Name of the fuzzer to which this model belongs.

  Returns:
    Model path. For example, if `/tmp/model` is the directory containing model
    files(e.g. rnn.index), the path should be '/tmp/model/rnn'.
  """
  # Get temporary directory.
  temp_directory = environment.get_value('BOT_TMPDIR')

  # Create model directory.
  model_directory = os.path.join(temp_directory, fuzzer_name)
  shell.remove_directory(model_directory, recreate=True)

  if not download_model_from_gcs(model_directory, fuzzer_name):
    return None

  # Got the model. Return model path.
  return os.path.join(model_directory, constants.RNN_MODEL_NAME)


def run(input_directory,
        output_directory,
        model_path,
        generation_timeout,
        generation_count=None,
        hidden_state_size=None,
        hidden_layer_size=None):
  """Generate inputs with specified model parameters.

  Args:
    input_directory: Corpus directory. Required argument for generation script.
    output_directory: New inputs directory. Required argument for generation
        script.
    model_path: Model path. Required argument for generation script.
    generation_timeout: Timeout for running generation process.
    generation_count: Number of inputs to generate. Required argument for
        generation script.
    hidden_state_size: Hidden state size of LSTM cell.
    hidden_layer_size: Hidden layer size of LSTM model.

  Returns:
    Result of running generation process. Format is defined by
    ProcessRunner.run_and_wait().
  """
  # Get generation script path.
  script_path = os.path.join(ML_RNN_SCRIPT_DIR,
                             constants.GENERATION_SCRIPT_NAME)

  # Wrap command arguments.
  args_list = [
      script_path,
      constants.INPUT_DIR_ARGUMENT_PREFIX + input_directory,
      constants.OUTPUT_DIR_ARGUMENT_PREFIX + output_directory,
      constants.MODEL_PATH_ARGUMENT_PREFIX + model_path,
  ]

  if generation_count:
    args_list.append(constants.GENERATION_COUNT_ARGUMENT_PREFIX +
                     str(generation_count))
  else:
    args_list.append(constants.GENERATION_COUNT_ARGUMENT_PREFIX +
                     str(GENERATION_MAX_COUNT))

  # Optional arguments.
  if hidden_state_size:
    args_list.append(constants.HIDDEN_STATE_ARGUMENT_PREFIX +
                     str(hidden_state_size))
  if hidden_layer_size:
    args_list.append(constants.HIDDEN_LAYER_ARGUMENT_PREFIX +
                     str(hidden_layer_size))

  script_environment = os.environ.copy()

  # Run process in script directory.
  rnn_runner = new_process.ProcessRunner(sys.executable)
  return rnn_runner.run_and_wait(
      args_list,
      cwd=ML_RNN_SCRIPT_DIR,
      env=script_environment,
      timeout=generation_timeout)


def execute(input_directory, output_directory, fuzzer_name, generation_timeout):
  """Execute ML RNN generator to produce new inputs.

  This method should be called inside launcher, to generate a number of
  new inputs based on ML RNN model.

  It will fetch ML model from GCS bucket specified in environment
  variable `CORPUS_BUCKET`. The script to run the model resides
  in folder `tools/fuzzers/ml/rnn`.

  Args:
    input_directory: Seed corpus path. The directory should not be empty.
    output_directory: The directory to place generated inputs.
    fuzzer_name: Name of the fuzzer, e.g libpng_read_fuzzer. It indicates the
        subdirectory in gcs bucket to store models.
    generation_timeout: Time in seconds for the generator to run. Normally it
        takes <1s to generate an input, assuming the input length is <4KB.
  """
  if environment.platform() != 'LINUX':
    logs.log('Unsupported platform for ML RNN generation, skipping.')
    return

  # Validate corpus folder.
  file_count = shell.get_directory_file_count(input_directory)
  if not file_count:
    logs.log('Corpus is empty. Skip generation.')
    return

  # Number of existing new inputs. They are possibly generated by other
  # generators.
  old_corpus_units = shell.get_directory_file_count(output_directory)
  old_corpus_bytes = shell.get_directory_size(output_directory)

  # Get model path.
  model_path = prepare_model_directory(fuzzer_name)
  if not model_path:
    return

  result = run(input_directory, output_directory, model_path,
               generation_timeout)

  # Generation process exited abnormally but not caused by timeout, meaning
  # error occurred during execution.
  if result.return_code and not result.timed_out:
    if result.return_code == constants.ExitCode.CORPUS_TOO_SMALL:
      logs.log_warn(
          'ML RNN generation for fuzzer %s aborted due to small corpus.' %
          fuzzer_name)
    else:
      logs.log_error(
          'ML RNN generation for fuzzer %s failed with ExitCode = %d.' %
          (fuzzer_name, result.return_code),
          output=utils.decode_to_unicode(result.output))
    return

  # Timeout is not error, if we have new units generated.
  if result.timed_out:
    logs.log_warn('ML RNN generation for fuzzer %s timed out.' % fuzzer_name)

  new_corpus_units = (
      shell.get_directory_file_count(output_directory) - old_corpus_units)
  new_corpus_bytes = (
      shell.get_directory_size(output_directory) - old_corpus_bytes)
  if new_corpus_units:
    logs.log('Added %d new inputs (%d bytes) using ML RNN generator for %s.' %
             (new_corpus_units, new_corpus_bytes, fuzzer_name))
  else:
    logs.log_error(
        'ML RNN generator did not produce any inputs for %s' % fuzzer_name,
        output=utils.decode_to_unicode(result.output))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests fuzzers.builtin."""

import os
import unittest

import parameterized
from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.bot.fuzzers import builtin
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class TestEngineFuzzer(builtin.EngineFuzzer):
  """A test engine fuzzer."""

  def generate_arguments(self, *_):  # pylint: disable=arguments-differ
    return '-arg1 -arg2'


class BaseEngineFuzzerTest(fake_filesystem_unittest.TestCase):
  """Engine fuzzer tests."""

  def setUp(self):
    """Setup for base engine fuzzer test."""
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.default_project_name',
        'clusterfuzz._internal.bot.fuzzers.builtin.fuzzers_utils.get_fuzz_targets'
    ])

    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir('/input')
    self.fs.create_dir('/output')

    environment.set_value('BUILD_DIR', '/build_dir')
    environment.set_value('FAIL_RETRIES', 1)

    environment.set_value('PROJECT_NAME', 'proj')
    self.mock.default_project_name.return_value = 'default-proj'

    self.mock.get_fuzz_targets.return_value = [
        '/build_dir/target',
    ]
    self.fs.create_file(
        '/build_dir/target.owners',
        contents='dev1@example1.com\ndev2@example2.com')


class EngineFuzzerTest(BaseEngineFuzzerTest):
  """Engine fuzzer tests."""

  def test_run(self):
    """Test running an engine fuzzer."""
    fuzzer = TestEngineFuzzer()
    result = fuzzer.run('/input', '/output', 1)

    self.assertEqual(
        'Generated 1 testcase for fuzzer target.\n'
        'metadata::fuzzer_binary_name: target\n'
        'metadata::issue_owners: dev1@example1.com,dev2@example2.com\n',
        result.output)
    self.assertEqual('/input/proj_target', result.corpus_directory)

    self.assertTrue(os.path.exists('/output/fuzz-0'))
    self.assertTrue(os.path.exists('/output/flags-0'))

    with open('/output/fuzz-0') as f:
      self.assertEqual(' ', f.read())

    with open('/output/flags-0') as f:
      self.assertEqual('%TESTCASE% target -arg1 -arg2', f.read())

  def test_run_with_labels(self):
    """Test running an engine fuzzer with a labels file."""
    self.fs.create_file('/build_dir/target.labels', contents='label1\nlabel2\n')

    fuzzer = TestEngineFuzzer()
    result = fuzzer.run('/input', '/output', 1)
    self.assertEqual(
        'Generated 1 testcase for fuzzer target.\n'
        'metadata::fuzzer_binary_name: target\n'
        'metadata::issue_owners: dev1@example1.com,dev2@example2.com\n'
        'metadata::issue_labels: label1,label2\n', result.output)

  def test_run_no_build_dir(self):
    """Test running without a build dir."""
    environment.set_value('BUILD_DIR', '')
    fuzzer = TestEngineFuzzer()
    with self.assertRaisesRegex(builtin.BuiltinFuzzerException, 'BUILD_DIR'):
      fuzzer.run('/input', '/output', 1)

  def test_run_no_fuzzers(self):
    """Test running without fuzzers."""
    self.mock.get_fuzz_targets.return_value = []
    fuzzer = TestEngineFuzzer()
    with self.assertRaises(builtin.BuiltinFuzzerException):
      fuzzer.run('/input', '/output', 1)

  def _generate_targets_list(self, count):
    """Generate a targets list."""
    fake_targets_list = []
    for i in range(count):
      fake_targets_list.append('/build_dir/target' + str(i))

    return fake_targets_list

  def test_run_chosen_fuzz_target(self):
    """Test running with chosen fuzz target."""
    os.environ['FUZZ_TARGET'] = 'chosen_target'

    fake_targets_list = self._generate_targets_list(100)
    fake_targets_list.append('/build_dir/chosen_target')

    self.mock.get_fuzz_targets.return_value = fake_targets_list

    fuzzer = TestEngineFuzzer()
    result = fuzzer.run('/input', '/output', 1)

    self.assertEqual(
        'Generated 1 testcase for fuzzer chosen_target.\n'
        'metadata::fuzzer_binary_name: chosen_target\n', result.output)
    self.assertEqual('/input/proj_chosen_target', result.corpus_directory)

  def test_sanitizer_options_from_options_file(self):
    """Tests that sanitizer options are set in *SAN_OPTIONS using the overrides
    provided in .options file."""
    environment.set_value('ASAN_OPTIONS', 'fake_option1=1')
    with open('/build_dir/target.options', 'w') as f:
      f.write('[asan]\nfake_option2=1\n[msan]\nfake_options3=1')

    fuzzer = TestEngineFuzzer()
    fuzzer.run('/input', '/output', 1)

    self.assertEqual('fake_option1=1:fake_option2=1',
                     environment.get_value('ASAN_OPTIONS'))
    self.assertEqual(None, environment.get_value('MSAN_OPTIONS'))


class GetFuzzerPath(unittest.TestCase):
  """_get_fuzzer_path Tests."""

  @parameterized.parameterized.expand([('fuzzer', 'LINUX'), ('fuzzer.exe',
                                                             'WINDOWS')])
  def test_get_fuzzer_path(self, target_name, mock_platform):
    """Test that get_fuzzer_path returns the path of a fuzzer."""
    target_path = os.path.join('path', 'to', target_name)
    helpers.patch(self, ['clusterfuzz._internal.system.environment.platform'])
    self.mock.platform.return_value = mock_platform
    result = builtin._get_fuzzer_path(['a', target_path], 'fuzzer')  # pylint: disable=protected-access
    self.assertEqual(result, target_path)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage fuzzers types."""

import datetime
import io

from flask import request
from google.cloud import ndb

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.fuzzing import fuzzer_selection
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import fuzzer_logs
from clusterfuzz._internal.system import archive
from handlers import base_handler
from libs import access
from libs import form
from libs import gcs
from libs import handler
from libs import helpers

ARCHIVE_READ_SIZE_LIMIT = 16 * 1024 * 1024


class Handler(base_handler.Handler):
  """Manages fuzzers."""

  @handler.get(handler.HTML)
  @handler.check_admin_access_if_oss_fuzz
  @handler.check_user_access(need_privileged_access=False)
  def get(self):
    """Handle a get request."""
    fuzzer_logs_bucket = fuzzer_logs.get_bucket()
    fuzzers = list(data_types.Fuzzer.query().order(data_types.Fuzzer.name))
    jobs = data_handler.get_all_job_type_names()
    corpora = [
        bundle.name for bundle in data_types.DataBundle.query().order(
            data_types.DataBundle.name)
    ]

    privileged = access.has_access(need_privileged_access=True)
    # Unprivileged users can't download fuzzers, so hide the download keys.
    if not privileged:
      for fuzzer in fuzzers:
        fuzzer.blobstore_key = ''

    template_values = {
        'privileged': privileged,
        'fuzzers': fuzzers,
        'fuzzerLogsBucket': fuzzer_logs_bucket,
        'fieldValues': {
            'corpora': corpora,
            'jobs': jobs,
            'uploadInfo': gcs.prepare_blob_upload()._asdict(),
            'csrfToken': form.generate_csrf_token(),
        }
    }

    return self.render('fuzzers.html', template_values)


class BaseEditHandler(base_handler.GcsUploadHandler):
  """Base edit handler."""

  def _read_to_bytesio(self, gcs_path):
    """Return a bytesio representing a GCS object."""
    data = storage.read_data(gcs_path)
    if not data:
      raise helpers.EarlyExitException('Failed to read uploaded archive.', 500)

    return io.BytesIO(data)

  def _get_executable_path(self, upload_info):
    """Get executable path."""
    executable_path = request.get('executable_path')
    if not upload_info:
      return executable_path

    if upload_info.size > ARCHIVE_READ_SIZE_LIMIT:
      return executable_path

    if not executable_path:
      executable_path = 'run'  # Check for default.

    reader = self._read_to_bytesio(upload_info.gcs_path)
    return archive.get_first_file_matching(executable_path, reader,
                                           upload_info.filename)

  def _get_launcher_script(self, upload_info):
    """Get launcher script path."""
    launcher_script = request.get('launcher_script')
    if not upload_info:
      return launcher_script

    if not launcher_script:
      return None

    if upload_info.size > ARCHIVE_READ_SIZE_LIMIT:
      return launcher_script

    reader = self._read_to_bytesio(upload_info.gcs_path)
    launcher_script = archive.get_first_file_matching(launcher_script, reader,
                                                      upload_info.filename)
    if not launcher_script:
      raise helpers.EarlyExitException(
          'Specified launcher script was not found in archive!', 400)

    return launcher_script

  def _get_integer_value(self, key):
    """Check a numeric input value."""
    value = request.get(key)
    if value is None:
      return None

    try:
      value = int(value)
    except (ValueError, TypeError):
      raise helpers.EarlyExitException(
          '{key} must be an integer.'.format(key=key), 400)

    if value <= 0:
      raise helpers.EarlyExitException(
          '{key} must be > 0.'.format(key=key), 400)

    return value

  def apply_fuzzer_changes(self, fuzzer, upload_info):
    """Apply changes to a fuzzer."""
    if upload_info and not archive.is_archive(upload_info.filename):
      raise helpers.EarlyExitException(
          'Sorry, only zip, tgz, tar.gz, tbz, and tar.bz2 archives are '
          'allowed!', 400)

    if fuzzer.builtin:
      executable_path = launcher_script = None
    else:
      executable_path = self._get_executable_path(upload_info)
      launcher_script = self._get_launcher_script(upload_info)

      # Executable path is required for non-builtin fuzzers and if it is not
      # already set.
      if not fuzzer.executable_path and not executable_path:
        raise helpers.EarlyExitException(
            'Please enter the path to the executable, or if the archive you '
            'uploaded is less than 16MB, ensure that the executable file has '
            '"run" in its name.', 400)

    jobs = request.get('jobs', [])
    timeout = self._get_integer_value('timeout')
    max_testcases = self._get_integer_value('max_testcases')
    external_contribution = request.get('external_contribution', False)
    differential = request.get('differential', False)
    environment_string = request.get('additional_environment_string')
    data_bundle_name = request.get('data_bundle_name')

    # Save the fuzzer file metadata.
    if upload_info:
      fuzzer.filename = upload_info.filename
      fuzzer.blobstore_key = str(upload_info.key())
      fuzzer.file_size = utils.get_size_string(upload_info.size)

    fuzzer.jobs = jobs
    fuzzer.revision = fuzzer.revision + 1
    fuzzer.source = helpers.get_user_email()
    fuzzer.timeout = timeout
    fuzzer.max_testcases = max_testcases
    fuzzer.result = None
    fuzzer.sample_testcase = None
    fuzzer.console_output = None
    fuzzer.external_contribution = bool(external_contribution)
    fuzzer.differential = bool(differential)
    fuzzer.additional_environment_string = environment_string
    fuzzer.timestamp = datetime.datetime.utcnow()
    fuzzer.data_bundle_name = data_bundle_name

    # Update only if a new archive is provided.
    if executable_path:
      fuzzer.executable_path = executable_path

    # Optional. Also, update only if a new archive is provided and contains a
    # launcher script.
    if launcher_script:
      fuzzer.launcher_script = launcher_script

    fuzzer.put()

    fuzzer_selection.update_mappings_for_fuzzer(fuzzer)

    helpers.log('Uploaded fuzzer %s.' % fuzzer.name, helpers.MODIFY_OPERATION)
    return self.redirect('/fuzzers')


class CreateHandler(BaseEditHandler):
  """Create a new fuzzer."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.check_user_access(need_privileged_access=True)
  @handler.require_csrf_token
  def post(self):
    """Handle a post request."""
    name = request.get('name')
    if not name:
      raise helpers.EarlyExitException('Please give the fuzzer a name!', 400)

    if not data_types.Fuzzer.VALID_NAME_REGEX.match(name):
      raise helpers.EarlyExitException(
          'Fuzzer name can only contain letters, numbers, dashes and '
          'underscores.', 400)

    existing_fuzzer = data_types.Fuzzer.query(data_types.Fuzzer.name == name)
    if existing_fuzzer.get():
      raise helpers.EarlyExitException(
          'Fuzzer already exists. Please use the EDIT button for changes.', 400)

    upload_info = self.get_upload()
    if not upload_info:
      raise helpers.EarlyExitException('Need to upload an archive.', 400)

    fuzzer = data_types.Fuzzer()
    fuzzer.name = name
    fuzzer.revision = 0
    return self.apply_fuzzer_changes(fuzzer, upload_info)


class EditHandler(BaseEditHandler):
  """Edit or create a fuzzer."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.check_user_access(need_privileged_access=True)
  @handler.require_csrf_token
  def post(self):
    """Handle a post request."""
    key = helpers.get_integer_key(request)

    fuzzer = ndb.Key(data_types.Fuzzer, key).get()
    if not fuzzer:
      raise helpers.EarlyExitException('Fuzzer not found.', 400)

    upload_info = self.get_upload()
    return self.apply_fuzzer_changes(fuzzer, upload_info)


class DeleteHandler(base_handler.Handler):
  """Delete a fuzzer."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.check_user_access(need_privileged_access=True)
  @handler.require_csrf_token
  def post(self):
    """Handle a post request."""
    key = helpers.get_integer_key(request)

    fuzzer = ndb.Key(data_types.Fuzzer, key).get()
    if not fuzzer:
      raise helpers.EarlyExitException('Fuzzer not found.', 400)

    fuzzer_selection.update_mappings_for_fuzzer(fuzzer, mappings=[])
    fuzzer.key.delete()

    helpers.log('Deleted fuzzer %s' % fuzzer.name, helpers.MODIFY_OPERATION)
    return self.redirect('/fuzzers')


class LogHandler(base_handler.Handler):
  """Show the console output from a fuzzer run."""

  @handler.check_user_access(need_privileged_access=False)
  def get(self, fuzzer_name):
    """Handle a get request."""
    helpers.log('LogHandler', fuzzer_name)
    fuzzer = data_types.Fuzzer.query(
        data_types.Fuzzer.name == fuzzer_name).get()
    if not fuzzer:
      raise helpers.EarlyExitException('Fuzzer not found.', 400)

    return self.render('viewer.html', {
        'title': 'Output for ' + fuzzer.name,
        'content': fuzzer.console_output,
    })



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test the launcher.py script for AFL-based fuzzers."""
# pylint: disable=protected-access

from functools import partial
import os

import mock
from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers.afl import fuzzer
from clusterfuzz._internal.bot.fuzzers.afl import launcher
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.tests.core.bot.fuzzers.engine_common_test import \
    GetTimeoutTestBase
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class LauncherTestBase(fake_filesystem_unittest.TestCase):
  """Provides common functionality used by TestCases for launcher.py."""
  INPUT_DIR = '/inputdir'
  TARGET_PATH = '/target'
  TARGET_OPTIONS_PATH = TARGET_PATH + '.options'
  TEMP_DIR = '/tmp'
  BUILD_DIR = '/build'
  OUTPUT_DIR = '/tmp/afl_output_dir'
  CRASHES_DIR = '/tmp/afl_output_dir/default/crashes'
  DEFAULT_INPUT_DIR_CONTENTS = [fuzzer.AFL_DUMMY_INPUT]
  DATA_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')

  def setUp(self):
    """Setup for launcher test base."""
    test_helpers.patch_environ(self)
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['BUILD_DIR'] = self.BUILD_DIR

    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir(self.INPUT_DIR)
    if not os.path.exists(self.TEMP_DIR):
      self.fs.create_dir(self.TEMP_DIR)

    self._create_file(fuzzer.AFL_DUMMY_INPUT)
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.utils.get_temp_dir',
    ])

    self.mock.get_temp_dir.return_value = self.TEMP_DIR
    dont_use_strategies(self)

  def _create_file(self, filename, directory=None, contents=' '):
    """Write |contents| to |filename| in self.INPUT_DIR. Create |filename| if it
    doesn't exist.
    """
    if directory is None:
      directory = self.INPUT_DIR

    file_path = os.path.join(directory, filename)
    return file_path, self.fs.create_file(file_path, contents=contents)

  def _assert_elements_equal(self, l1, l2):
    """Assert that the elements of |l1| and |l2|. Modifies |l1| and |l2| by
    sorting them."""
    self.assertEqual(sorted(l1), sorted(l2))


class FuzzingStrategiesTest(fake_filesystem_unittest.TestCase):
  """Tests for launcher.FuzzingStrategies."""

  NUM_FILES = 1000
  INPUT_DIR = '/input'

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    for file_num in range(self.NUM_FILES):
      self.fs.create_file(os.path.join(self.INPUT_DIR, str(file_num)))

    test_helpers.patch(
        self,
        ['clusterfuzz._internal.bot.fuzzers.engine_common.is_lpm_fuzz_target'])
    self.mock.is_lpm_fuzz_target.return_value = True
    self.strategies = launcher.FuzzingStrategies(None)


class AflFuzzInputDirectoryTest(LauncherTestBase):
  """Test the launcher.AflFuzzInputDirectory class. Note that most of the
  methods tested are called from AflFuzzInputDirectory.__init__, so we will
  create the object rather than call them directly in these cases."""

  def setUp(self):
    super().setUp()
    self.temp_input_dir = os.path.join(self.TEMP_DIR, 'afl_input_dir')
    self.fs.create_dir(self.temp_input_dir)
    test_helpers.patch(
        self,
        ['clusterfuzz._internal.bot.fuzzers.engine_common.is_lpm_fuzz_target'])
    self.mock.is_lpm_fuzz_target.return_value = True
    self.strategies = launcher.FuzzingStrategies(None)

  def _new_afl_input(self):
    """Create a new AflFuzzInputDirectory object."""
    return launcher.AflFuzzInputDirectory(self.temp_input_dir, self.TARGET_PATH,
                                          self.strategies)

  def test_corpus_subset(self):
    """Tests that create_new_if_needed works as intended when told to use a 75
    file corpus subset."""
    self.strategies.use_corpus_subset = True
    self.strategies.corpus_subset_size = 75
    # Now test create_new_if_needed obeys corpus_subset.
    self.strategies.use_corpus_subset = True
    for file_num in range(self.strategies.corpus_subset_size):
      self._create_file(str(file_num), directory=self.temp_input_dir)

    afl_input = self._new_afl_input()
    self.assertEqual(
        len(os.listdir(afl_input.input_directory)),
        self.strategies.corpus_subset_size)

    self.assertTrue(self.strategies.use_corpus_subset)


class AflFuzzOutputDirectoryTest(LauncherTestBase):
  """Test the launcher.AflFuzzOutputDirectory class."""
  QUEUE_DIR = '/tmp/afl_output_dir/default/queue'

  # Note that this is just a file that is part of the input corpus.
  # It isn't an AFL testcase.
  INPUT_TESTCASE_FILENAME = '1'
  INPUT_TESTCASE_INO = 1
  QUEUE_INPUT_LINK_FILENAME = 'id:000000,orig:1'
  QUEUE_INPUT_COPIED_FILENAME = 'id:000002,orig:2'
  QUEUE_TESTCASE_FILENAME = 'id:000001,src:000000,op:flip1,pos:0,+cov'
  QUEUE_TESTCASE_INO = 2

  def setUp(self):
    super().setUp()
    # Creates self.OUTPUT_DIR.
    self.afl_output = launcher.AflFuzzOutputDirectory()
    os.makedirs(self.CRASHES_DIR)
    os.makedirs(self.QUEUE_DIR)

    self.input_testcase_path, input_testcase_obj = self._create_file(
        self.INPUT_TESTCASE_FILENAME)

    input_testcase_obj.SetIno(self.INPUT_TESTCASE_INO)
    self.input_directory_inodes = set(
        [os.stat(self.input_testcase_path).st_ino])

    self.queue_input_link_path, queue_input_link_obj = self._create_file(
        self.QUEUE_INPUT_LINK_FILENAME, directory=self.QUEUE_DIR)

    queue_input_link_obj.SetIno(self.INPUT_TESTCASE_INO)

    self.queue_copied_testcase_path, _ = self._create_file(
        self.QUEUE_INPUT_COPIED_FILENAME, directory=self.QUEUE_DIR)

    self.queue_testcase_path, queue_testcase_obj = self._create_file(
        self.QUEUE_TESTCASE_FILENAME, directory=self.QUEUE_DIR)

    queue_testcase_obj.SetIno(self.QUEUE_TESTCASE_INO)

  def test_is_testcase(self):
    """Test that is_testcase() works as expected."""

    # Test that testcases must be files and not directories.
    dir_name = 'id:000004,src:000000,op:flip1,pos:0,+cov'
    os.mkdir(dir_name)
    self.assertFalse(self.afl_output.is_testcase(dir_name))

    # Test that it returns False for a non-testcase.
    non_testcase_filename = 'non-testcase'
    non_testcase_path, _ = self._create_file(non_testcase_filename)
    self.assertFalse(self.afl_output.is_testcase(non_testcase_path))

    # Test that files produced by AFL are considered testcases.
    self.assertTrue(self.afl_output.is_testcase(self.queue_testcase_path))

  def test_is_new_testcase(self):
    """Test that is_new_testcase behaves as expected when passed non-hard
    links."""
    # Test that it returns False for a non_testcase.
    non_testcase_filename = 'non-testcase'
    non_testcase_path, _ = self._create_file(non_testcase_filename)
    self.assertFalse(self.afl_output.is_new_testcase(non_testcase_path))

    # Test that it returns True for a new testcase.
    self.assertTrue(self.afl_output.is_new_testcase(self.queue_testcase_path))

    # Test that it returns False for a copied testcase.
    self.assertFalse(
        self.afl_output.is_new_testcase(self.queue_copied_testcase_path))

  def test_count_new_units(self):
    """Test that count_new_units works as expected."""
    self.afl_output.destination_directory_inodes = set()
    # Test that it works as intended.
    self.assertEqual(self.afl_output.count_new_units(self.afl_output.queue), 1)

    # Test that hard links aren't counted as new units because they are copies
    # of files from the original output directory.
    self.afl_output.destination_directory_inodes = self.input_directory_inodes
    self.assertEqual(self.afl_output.count_new_units(self.afl_output.queue), 1)

    # Test that copies aren't counted as new units.
    self.fs.RemoveObject(self.queue_copied_testcase_path)
    self.assertEqual(self.afl_output.count_new_units(self.afl_output.queue), 1)

    # Test that non-testcases aren't counted as new units.
    self._create_file('README.txt', directory=self.afl_output.queue)
    self.assertEqual(self.afl_output.count_new_units(self.afl_output.queue), 1)

  def test_remove_hang_in_queue(self):
    """Test that remove_hang_in_queue works as expected."""
    self.assertIn(self.QUEUE_INPUT_LINK_FILENAME, os.listdir(self.QUEUE_DIR))
    self.afl_output.remove_hang_in_queue(self.queue_input_link_path)
    self.assertNotIn(self.QUEUE_INPUT_LINK_FILENAME, os.listdir(self.QUEUE_DIR))


class SetSanitizerOptionsTest(LauncherTestBase):
  """Test that we set the sanitzer options correctly. This is very important
  as afl-fuzz will refuse to run if we do it incorrectly.
  """

  DUMMY_OPTION = 'dummy_option=dummy_value'
  REQUIRED_ASAN_OPTIONS = ['symbolize=0', 'abort_on_error=1']
  REQUIRED_MSAN_OPTIONS = ['symbolize=0', 'exit_code=86']

  def _del_opts(self):
    if 'ASAN_OPTIONS' in os.environ:
      del os.environ['ASAN_OPTIONS']

    if 'MSAN_OPTIONS' in os.environ:
      del os.environ['MSAN_OPTIONS']

  def assert_sanitizer_opts_set(self, sanitizer_options_variable,
                                *additional_options):
    """Test that sanitizer options are set."""
    if sanitizer_options_variable == 'ASAN_OPTIONS':
      required_options = self.REQUIRED_ASAN_OPTIONS
    else:
      required_options = self.REQUIRED_MSAN_OPTIONS

    if additional_options:
      opts = required_options + list(additional_options)
    else:
      opts = required_options

    sanitizer_options_value = os.environ[sanitizer_options_variable]
    for opt in opts:
      self.assertIn(opt, sanitizer_options_value)

  def setUp(self):
    super().setUp()
    test_helpers.patch_environ(self)

  def test_left_unset(self):
    launcher.set_additional_sanitizer_options_for_afl_fuzz()
    self.assertNotIn('ASAN_OPTIONS', os.environ)
    self.assertNotIn('MSAN_OPTIONS', os.environ)

  def test_set_when_empty(self):
    os.environ['ASAN_OPTIONS'] = ''
    os.environ['MSAN_OPTIONS'] = ''
    launcher.set_additional_sanitizer_options_for_afl_fuzz()
    self.assert_sanitizer_opts_set('ASAN_OPTIONS')
    self.assert_sanitizer_opts_set('MSAN_OPTIONS')

  def test_opts_preserved(self):
    """Test opts preserved."""
    os.environ['ASAN_OPTIONS'] = self.DUMMY_OPTION
    os.environ['MSAN_OPTIONS'] = self.DUMMY_OPTION
    launcher.set_additional_sanitizer_options_for_afl_fuzz()
    self.assert_sanitizer_opts_set('ASAN_OPTIONS', self.DUMMY_OPTION)
    self.assert_sanitizer_opts_set('MSAN_OPTIONS', self.DUMMY_OPTION)

    self._del_opts()
    os.environ['ASAN_OPTIONS'] = self.DUMMY_OPTION
    launcher.set_additional_sanitizer_options_for_afl_fuzz()
    self.assert_sanitizer_opts_set('ASAN_OPTIONS', self.DUMMY_OPTION)
    self.assertNotIn('MSAN_OPTIONS', os.environ)

  def test_options_file(self):
    """Test *SAN_OPTIONS set from .options file."""
    os.environ['ASAN_OPTIONS'] = self.DUMMY_OPTION
    os.environ['MSAN_OPTIONS'] = self.DUMMY_OPTION
    self._create_file(
        self.TARGET_OPTIONS_PATH,
        contents='[asan]\nfake_option=1\n[msan]\nfake_option=2')
    engine_common.process_sanitizer_options_overrides(self.TARGET_PATH)
    launcher.set_additional_sanitizer_options_for_afl_fuzz()
    self.assert_sanitizer_opts_set('ASAN_OPTIONS', self.DUMMY_OPTION,
                                   'fake_option=1')
    self.assert_sanitizer_opts_set('MSAN_OPTIONS', self.DUMMY_OPTION,
                                   'fake_option=2')


class AflRunnerTest(LauncherTestBase):
  """Tests for AflRunnerCommon."""
  TESTCASE_FILE_PATH = '/testcase'
  ARBITRARY_OUTPUT = 'No new instrumentation output, test case may be useless'
  ARBITRARY_RET_CODE = 99
  BAD_FILENAME = 'badfile'
  BAD_FILE_CONTENTS = 'bad'

  def setUp(self):
    super().setUp()
    test_helpers.patch_environ(self)
    test_helpers.patch(
        self,
        ['clusterfuzz._internal.bot.fuzzers.engine_common.is_lpm_fuzz_target'])
    self.mock.is_lpm_fuzz_target.return_value = True
    environment.set_value('HARD_TIMEOUT_OVERRIDE', 600)
    config = launcher.AflConfig.from_target_path(self.TARGET_PATH)

    self.runner = launcher.AflRunner(self.TARGET_PATH, config,
                                     self.TESTCASE_FILE_PATH, self.INPUT_DIR)

    self.fuzz_result = new_process.ProcessResult()
    self.args = ['-iinput1', '-ooutput', '123', '456']
    self.times_called = 0

  def test_gen_afl_args(self):
    """Test that we are generating arguments for AFL correctly."""
    target_path = '/targetpath'
    output_dir = '/afl_output_dir'
    input_dir = '/inputdir'
    extra_args = ['extra1', 'extra2']
    testcase_file_path = '/testcase'
    config = launcher.AflConfig.from_target_path(target_path)
    config.additional_afl_arguments.extend(extra_args)
    runner = launcher.AflRunner(target_path, config, testcase_file_path,
                                input_dir)

    afl_args = runner.generate_afl_args()

    self.assertIn(target_path, afl_args)
    target_idx = afl_args.index(target_path)

    # Test that output_dir is the parameter for -o and that it is being passed
    # to afl and not the target binary.
    output_dir_argument = [arg for arg in afl_args if output_dir in arg][0]
    self.assertIn('-o', output_dir_argument)
    self.assertLess(afl_args.index(output_dir_argument), target_idx)

    input_dir_argument = [arg for arg in afl_args if input_dir in arg][0]
    self.assertIn('-i', input_dir_argument)
    self.assertLess(afl_args.index(input_dir_argument), target_idx)

    for extra_arg in extra_args:
      self.assertIn(extra_arg, afl_args)
      self.assertLess(afl_args.index(extra_arg), len(afl_args))

  def test_afl_setup(self):
    """Test AflRunner.afl_setup."""
    self._create_file(self.runner.stderr_file_path)
    self.runner.afl_setup()
    self.assertFalse(os.path.exists(self.runner.stderr_file_path))

  def test_set_environment_variables(self):
    """Test AflRunner.set_environment_variables."""
    self.runner.afl_setup()
    self.assertEqual(os.environ['AFL_DRIVER_STDERR_DUPLICATE_FILENAME'],
                     self.runner.stderr_file_path)
    self.assertEqual(os.environ['AFL_BENCH_UNTIL_CRASH'], '1')

  def test_set_resume(self):
    """Test AflRunner.set_resume."""
    self.runner.set_resume(self.args)
    self.assertIn('-i-', self.args)

  def test_get_arg_index(self):
    """Test AflRunner.get_arg_index."""
    # Test that it finds the correct index.
    self.assertEqual(self.runner.get_arg_index(self.args, '-o'), 1)
    # Test that it returns negative one when the flag is not self.args.
    self.assertEqual(self.runner.get_arg_index(self.args, '-z'), -1)

  def test_set_input_arg(self):
    """Test AflRunner.set_input_arg."""
    self.runner.set_input_arg(self.args, 'input2')
    self.assertEqual(self.args, ['-iinput2', '-ooutput', '123', '456'])

  def test_set_timeout_arg(self):
    """Test AflRunner.set_input_arg."""
    # Test that it is added when it doesn't yet exist.
    self.runner.set_timeout_arg(self.args, 1000)
    self.assertIn('-t1000', self.args)

    # Test that it is changed if it already exists.
    self.runner.set_timeout_arg(self.args, 2000)
    self.assertNotIn('-t1000', self.args)
    self.assertIn('-t2000', self.args)

    # Test that skip_hangs functionality works.
    self.runner.set_timeout_arg(self.args, 3000, True)
    self.assertIn('-t3000+', self.args)

  def test_set_arg(self):
    """Test AflRunnerCommon.set_arg."""
    self.runner.set_arg(self.args, '-i', 'input2')
    self.assertEqual(self.args, ['-iinput2', '-ooutput', '123', '456'])

    self.runner.set_arg(self.args, '-x', 'dict')
    self.assertEqual(self.args,
                     ['-xdict', '-iinput2', '-ooutput', '123', '456'])

  def test_should_try_fuzzing(self):
    """Test AflRunnerCommon.should_try_fuzzing."""
    # Test that it returns True when we haven't fuzzed yet.
    time_left = 100
    num_retries = 0
    self.assertTrue(self.runner.should_try_fuzzing(time_left, num_retries))

    time_left = 0
    # Test that it returns False when there is no time left to fuzz.
    self.assertFalse(self.runner.should_try_fuzzing(time_left, num_retries))

    time_left = 100
    num_retries = launcher.AflRunnerCommon.MAX_FUZZ_RETRIES + 1
    # Test that it returns False when there is no time left to fuzz.
    self.assertFalse(self.runner.should_try_fuzzing(time_left, num_retries))

  def test_prepare_retry_if_cpu_error(self):
    """Test AflRunner.prepare_retry_if_cpu_error."""
    # Test that it doesn't prepare a retry if the error is unrelated:
    if 'AFL_NO_AFFINITY' in os.environ:
      del os.environ['AFL_NO_AFFINITY']

    self.fuzz_result.output = (
        '[-] PROGRAM ABORT : All test cases time out, giving up!\n'
        'Location : perform_dry_run(), afl-fuzz.c:3240\n')

    self.assertFalse(self.runner.prepare_retry_if_cpu_error(self.fuzz_result))
    self.assertNotIn('AFL_NO_AFFINITY', os.environ)

    # Now test that it tries to fix the issue it is supposed to.
    self.fuzz_result.output = (
        '[-] PROGRAM ABORT : No more free CPU cores\n'
        'Location : bind_to_free_cpu(), afl-fuzz.c:484\n')

    self.assertTrue(self.runner.prepare_retry_if_cpu_error(self.fuzz_result))
    self.runner.afl_setup()
    self.assertIn('AFL_NO_AFFINITY', os.environ)
    self.assertEqual(os.environ['AFL_NO_AFFINITY'], '1')

    # Now test it doesn't try to fix the issue again (this is a sanity check
    # on our code).
    self.assertFalse(self.runner.prepare_retry_if_cpu_error(self.fuzz_result))

  def test_fuzzer_stderr_ioerror(self):
    """Test AflRunner.fuzzer_stderr when there is an error reading the
    stderr file."""
    test_helpers.patch(
        self, ['clusterfuzz._internal.base.utils.read_from_handle_truncated'])

    self.mock.read_from_handle_truncated.side_effect = IOError
    self.assertIsNone(self.runner._fuzzer_stderr)
    self.assertEqual(self.runner.fuzzer_stderr, '')

  def test_fuzzer_stderr(self):
    """Test AflRunner.fuzzer_stderr works correctly."""
    stderr = 'hello'
    self.fs.create_file(self.runner.stderr_file_path, contents=stderr)
    self.assertIsNone(self.runner._fuzzer_stderr)
    self.assertEqual(self.runner.fuzzer_stderr, stderr)

    # Now test that it doesn't read more than launcher.MAX_OUTPUT_LEN
    stderr = (launcher.MAX_OUTPUT_LEN + 1) * 'a'
    with open(self.runner.stderr_file_path, 'w+') as file_handle:
      file_handle.write(stderr)

    # A truncated marker will be added, so _fuzzer_stderr will actually be
    # allowed to be greater than MAX_OUTPUT_LEN, but not by much.
    self.assertLess(
        len(self.runner.fuzzer_stderr), launcher.MAX_OUTPUT_LEN + 50)

  def _process_result(self,
                      command=None,
                      output=None,
                      return_code=0,
                      time_executed=20,
                      timed_out=False):
    """Creates a new_process.ProcessResult with specified values or good
    defaults."""
    if command is None:
      command = ['afl-fuzz', '-iin', '-oout', './fuzzer']
    if output is None:
      output = self.ARBITRARY_OUTPUT

    return new_process.ProcessResult(
        command=command,
        output=output,
        return_code=return_code,
        time_executed=time_executed,
        timed_out=timed_out)

  def _initialization_for_run_afl(self):
    """Initialization."""
    # Test that it works when everything works normally (no errors).
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.afl.launcher.AflRunner.run_and_wait',
    ])

    # Make sure this initialized or else it will remove CRASHES_DIR.
    self.runner.afl_output  # pylint: disable=pointless-statement
    self.fs.create_dir(self.CRASHES_DIR)
    self.runner.initial_max_total_time = 100
    self.mock.run_and_wait.side_effect = (
        lambda *args, **kwargs: self._process_result())

    self._write_bad_input()

  def test_do_offline_mutations_small_testcase(self):
    """Tests that do_offline_mutations doesn't remove non-oversized testcases
    from the corpus."""
    # <1 MB testcase isn't oversized.
    self.assertTrue(self._do_offline_mutations(2**20 - 1))

  def test_do_offline_mutations_large_testcase(self):
    """Tests that do_offline_mutations doesn't add oversized testcases to the
    corpus."""
    # 1 MB testcase is oversized.
    self.assertFalse(self._do_offline_mutations(2**20))

  def _do_offline_mutations(self, size):
    """Creates a file |size| bytes long in the input directory, then calls
    do_offline_mutations and returns whether the file is in the input directory.
    """
    contents = 'A' * size  # 1 MB
    filename = 'test-file'
    input_dir = self.runner.afl_input.input_directory
    filepath = os.path.join(input_dir, filename)
    self.fs.create_file(filepath, contents=contents)
    self.runner.strategies.is_mutations_run = True
    self.runner.do_offline_mutations()
    return filename in os.listdir(input_dir)

  def test_run_afl_fuzz_fuzz_success(self):
    """Test AflRunner.run_afl_fuzz_and_handle_error when fuzzing succeeds."""
    self._initialization_for_run_afl()

    self.assertEqual(self.runner.run_afl_fuzz(self.args).return_code, 0)
    self.assertEqual(self.runner.fuzzer_stderr, '')

  def test_run_afl_fuzz_one_cpu_error(self):
    """Test AflRunner.run_afl_fuzz_and_handle_error works as intended when there
    is an error binding to CPU, but afl-fuzz is able to run in the end."""

    def one_cpu_error(*args, **kwargs):  # pylint: disable=unused-argument
      self.times_called += 1
      if self.times_called == 1:
        return self._process_result(
            output='PROGRAM ABORT :No more free CPU cores',
            return_code=self.ARBITRARY_RET_CODE)

      return self._process_result()

    self._initialization_for_run_afl()
    self.mock.run_and_wait.side_effect = one_cpu_error
    self.assertEqual(self.runner.run_afl_fuzz(self.args).return_code, 0)

  @mock.patch('clusterfuzz._internal.metrics.logs.log_error')
  def test_run_afl_fuzz_two_cpu_errors(self, mock_log_error):
    """Test AflRunner.run_afl_fuzz_and_handle_error works as intended when there
    is an error binding to CPU and afl-fuzz is never able to run in the end.
    Note that this should never happen in real life."""
    self._initialization_for_run_afl()

    def two_cpu_errors(*args, **kwargs):  # pylint: disable=unused-argument
      return self._process_result(
          output='PROGRAM ABORT :No more free CPU cores',
          return_code=self.ARBITRARY_RET_CODE)

    self.mock.run_and_wait.side_effect = two_cpu_errors
    fuzz_result = self.runner.run_afl_fuzz(self.args)
    self.assertNotEqual(0, fuzz_result.return_code)
    mock_log_error.assert_called_with(
        ('Afl exited with a non-zero exitcode: %s. Cannot recover.' %
         fuzz_result.return_code),
        engine_output=fuzz_result.output)

  def _write_bad_input(self):
    """Writes a "bad" file into the input directory."""
    self._create_file(self.BAD_FILENAME, contents=self.BAD_FILE_CONTENTS)


class GetFuzzTimeoutTest(GetTimeoutTestBase):
  """Get fuzz timeout tests."""
  function = staticmethod(
      partial(launcher.get_fuzz_timeout, is_mutations_run=False))

  def test_validation(self):
    """Test that get_fuzz_timeout rejects an invalid combination of
    HARD_TIMEOUT_OVERRIDE and MERGE_TIMEOUT_OVERRIDE."""
    self.validation_helper({
        'FUZZ_TEST_TIMEOUT': self.valid_hard_timeout,
        'HARD_TIMEOUT_OVERRIDE': self.valid_hard_timeout,
        'MERGE_TIMEOUT_OVERRIDE': self.valid_hard_timeout
    })

  def test_correctness(self):
    """Test that get_fuzz_timeout returns what we expect."""
    expected_fuzz_timeout = (
        self.valid_hard_timeout - launcher.POSTPROCESSING_TIMEOUT -
        self.valid_merge_timeout)

    self.call_helper(
        expected_fuzz_timeout, {
            'FUZZ_TEST_TIMEOUT': self.valid_hard_timeout,
            'MERGE_TIMEOUT_OVERRIDE': self.valid_merge_timeout
        })


class AflConfigTest(LauncherTestBase):
  """Test AflConfig."""

  def setUp(self):
    super().setUp()
    self.target_path = '/build_dir/target'
    self.options_path = self.target_path + '.options'

  def _create_options_file(self, contents):
    self._create_file(self.options_path, contents=contents)

  def _get_config(self):
    return launcher.AflConfig.from_target_path(self.target_path)

  def test_num_persistent_executions(self):
    """"Test that AflConfig sets the number of persistent executions we
    specified."""
    self._create_options_file('[libfuzzer]\n'
                              'dict=my_dict.dict\n'
                              'max_len=1337\n'
                              '[AFL]\n'
                              'N=800\n')

    self.assertEqual('800', self._get_config().num_persistent_executions)

  def test_dont_defer(self):
    """"Test that AflConfig allows .options file to opt-out of AFL's
    forkserver."""
    self._create_options_file('[env]\n'
                              'INVALID_ENV_VAR=blah\n'
                              'AFL_DRIVER_DONT_DEFER=1\n')

    env = self._get_config().additional_env_vars
    self.assertEqual('1', env.get('AFL_DRIVER_DONT_DEFER', None))

  def test_use_default_dict_without_fuzz_target_extension(self):
    """Test that AflConfig uses the default dict when none is specified. This
    is for case when fuzz target does not have an extension."""
    self.target_path = '/build_dir/target'
    dict_path = '/build_dir/target.dict'
    self._create_file(dict_path, contents='')

    self.assertIn('-x' + dict_path, self._get_config().additional_afl_arguments)

  def test_use_default_dict_with_fuzz_target_extension(self):
    """Test that AflConfig uses the default dict when none is specified. This
    is for the case when fuzz target has an extension."""
    self.target_path = '/build_dir/target.exe'
    dict_path = '/build_dir/target.dict'
    self._create_file(dict_path, contents='')

    self.assertIn('-x' + dict_path, self._get_config().additional_afl_arguments)

  def test_libfuzzer_section_dict(self):
    """"Test that AflConfig uses the dict specified by the libfuzzer section of
    a .options file."""

    self._create_options_file('[libfuzzer]\n'
                              'dict=my_dict.dict\n'
                              'max_len=1337\n')

    self.assertIn('-x/build_dir/my_dict.dict',
                  self._get_config().additional_afl_arguments)

  def test_close_fd_mask(self):
    """"Test that AflConfig instructs afl_driver to close fd mask if specified
    by the libfuzzer section of a .options file."""

    fd_mask_value = 3
    self._create_options_file(
        ('[libfuzzer]\n'
         'close_fd_mask={fd_mask_value}\n'
         'max_len=1337\n').format(fd_mask_value=fd_mask_value))

    afl_additional_env_vars = self._get_config().additional_env_vars
    self.assertIn('AFL_DRIVER_CLOSE_FD_MASK', afl_additional_env_vars)
    self.assertEqual(
        str(fd_mask_value), afl_additional_env_vars['AFL_DRIVER_CLOSE_FD_MASK'])


class ListFullFilePathsTest(LauncherTestBase):
  """Tests for list_full_file_paths."""
  DUMMY_2_FILENAME = 'dummyfile2'
  DUMMY_3_FILENAME = 'dummyfile3'

  def setUp(self):
    super().setUp()
    self.dummy_file_path = os.path.join(self.INPUT_DIR, fuzzer.AFL_DUMMY_INPUT)
    self.dummy_3_file_path, _ = self._create_file(self.DUMMY_3_FILENAME)

  def test_list_full_file_paths(self):
    """Test that list_full_file_paths works as intended."""
    # Test it works with just files:
    self._assert_elements_equal(
        launcher.list_full_file_paths(self.INPUT_DIR),
        [self.dummy_file_path, self.dummy_3_file_path])

    # Test it works with a subdirectory by not returning it:
    dummy_dir = os.path.join(self.INPUT_DIR, 'dummydir')
    self.fs.create_dir(dummy_dir)
    self._create_file(self.DUMMY_2_FILENAME, directory=dummy_dir)

    self._assert_elements_equal(
        launcher.list_full_file_paths(self.INPUT_DIR),
        [self.dummy_file_path, self.dummy_3_file_path])


class CorpusTest(fake_filesystem_unittest.TestCase):
  """Tests for Corpus and CorpusElement classes."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    self.corpus = launcher.Corpus()
    self.guard = 0

  def _get_unique_feature(self):
    """Returns an arbitrary, unique, feature for use in testing."""
    guard = self.guard
    self.guard += 1
    default_hit_count = 1
    return (guard, default_hit_count)

  def _create_file(self, path, size=1):
    """Creates a file at |path| that is |size| bytes large."""
    self.fs.create_file(path, contents=size * 'A')

  def test_corpus_element(self):
    """Tests CorpusElement class."""
    path = '/path/to/file'
    size = 20
    self._create_file(path, size=size)
    corpus_element = launcher.CorpusElement(path)
    self.assertEqual(path, corpus_element.path)
    self.assertEqual(size, corpus_element.size)

  def test_element_paths(self):
    """Tests that element_paths is the set of filepaths of elements in the
    corpus."""
    filenames = ['file_1', 'file_2']
    for filename in filenames:
      feature = self._get_unique_feature()
      self._create_file(filename)
      self.corpus.associate_features_with_file([feature], filename)
    self.assertEqual(set(filenames), self.corpus.element_paths)

  def test_associate_new_features_with_file(self):
    """Tests that associate_features_with_file associates new features with a
    file."""
    # Create an arbitrary number of features.
    features = [self._get_unique_feature() for _ in range(3)]
    filename = 'element'
    self._create_file(filename)
    self.corpus.associate_features_with_file(features, filename)
    for feature in features:
      self.assertEqual(filename,
                       self.corpus.features_and_elements[feature].path)

  def test_associate_feature_with_smaller_file(self):
    """Tests that associate_features_with_file associates features with the
    smallest file. Also verify that an element that isn't the smallest
    associated with any feature isn't part of the corpus."""
    features = [self._get_unique_feature()]
    larger_filename = 'larger'
    self._create_file(larger_filename, size=2)
    self.corpus.associate_features_with_file(features, larger_filename)
    smaller_filename = 'smaller'
    self._create_file(smaller_filename, size=1)
    self.corpus.associate_features_with_file(features, smaller_filename)
    self.assertEqual(smaller_filename,
                     self.corpus.features_and_elements[features[0]].path)
    self.assertEqual(set([smaller_filename]), self.corpus.element_paths)

  def test_file_with_one_feature_remains(self):
    """Test that a file remains in the corpus as long as it the smallest element
    for at least one feature."""
    feature_1 = self._get_unique_feature()
    feature_2 = self._get_unique_feature()
    larger_filename = 'larger'
    self._create_file(larger_filename, size=2)
    self.corpus.associate_features_with_file([feature_1, feature_2],
                                             larger_filename)
    smaller_filename = 'smaller'
    self._create_file(smaller_filename, size=1)
    self.corpus.associate_features_with_file([feature_2], smaller_filename)
    self.assertEqual(smaller_filename,
                     self.corpus.features_and_elements[feature_2].path)
    self.assertEqual(larger_filename,
                     self.corpus.features_and_elements[feature_1].path)
    self.assertEqual(
        set([smaller_filename, larger_filename]), self.corpus.element_paths)


def dont_use_strategies(obj):
  """Helper function to prevent using fuzzing strategies, unless asked for."""
  test_helpers.patch(obj, [
      'clusterfuzz._internal.bot.fuzzers.engine_common.decide_with_probability',
  ])
  obj.mock.decide_with_probability.return_value = False



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for init_runner."""

import unittest

from clusterfuzz._internal.bot.init_scripts import init_runner
from clusterfuzz._internal.tests.test_libs import helpers


class InitRunnerTest(unittest.TestCase):
  """Tests for init_runner."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.system.environment.platform',
        'clusterfuzz._internal.system.process_handler.run_process',
    ])

  def test_windows(self):
    """Test windows."""
    self.mock.platform.return_value = 'WINDOWS'
    init_runner.run()
    self.mock.run_process.assert_called_with(
        'powershell.exe ./configs/test/bot/init/windows.ps1',
        ignore_children=True,
        need_shell=True,
        testcase_run=False,
        timeout=1800)

  def test_posix(self):
    """Test posix."""
    self.mock.platform.return_value = 'LINUX'
    init_runner.run()
    self.mock.run_process.assert_called_with(
        './configs/test/bot/init/linux.bash',
        ignore_children=True,
        need_shell=True,
        testcase_run=False,
        timeout=1800)

  def test_nonexistent_platform(self):
    """Test posix."""
    self.mock.platform.return_value = 'FAKE'
    init_runner.run()
    self.assertEqual(0, self.mock.run_process.call_count)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the revisions module."""

import ast
import hashlib
import os
import unittest

import mock

from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import mock_config
from clusterfuzz._internal.tests.test_libs import test_utils

ANDROID_JOB_TYPE = 'android_asan_chrome'
BASIC_JOB_TYPE = 'linux_asan_chrome_mp'
SRCMAP_JOB_TYPE = 'linux_asan_libass'
CUSTOM_BINARY_JOB_TYPE = 'linux_asan_custom'
DATA_DIRECTORY = os.path.join(os.path.dirname(__file__), 'revisions_data')

# For simplicity, the mocked URLs all end with the revision. They do not match
# the ones in use in production exactly.
REVISION_VARS_URL = (
    'default;https://chromium.googlesource.com/'
    'chromium/src/+/%s/DEPS?format=text\n'
    'android_asan_chrome;https://commondatastorage.googleapis.com/'
    'chrome-test-builds/android/revisions/%s\n'
    'linux_asan_libass;'
    'https://commondatastorage.googleapis.com/blah-%s.srcmap.json\n'
    'v8;https://test-repository/src.git/+/%s?format=JSON')


@test_utils.with_cloud_emulators('datastore')
class RevisionsTestcase(unittest.TestCase):
  """Revisions tests."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.default_project_name',
        'clusterfuzz._internal.base.memoize.FifoOnDisk.get',
        'clusterfuzz._internal.base.memoize.FifoOnDisk.put',
        'clusterfuzz._internal.config.local_config.ProjectConfig',
    ])

    self.mock.get.return_value = None
    self.mock.default_project_name.return_value = 'project'
    self.mock.ProjectConfig.return_value = mock_config.MockConfig({
        'env': {
            'REVISION_VARS_URL':
                'https://chromium.googlesource.com/chromium/src/+/%s/DEPS'
                '?format=text'
        }
    })

    os.environ['REVISION_VARS_URL'] = (
        'https://chromium.googlesource.com/chromium/src/+/%s/DEPS?format=text')
    data_types.Job(
        name=ANDROID_JOB_TYPE,
        environment_string=(
            'HELP_URL = help_url\n'
            'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
            'chrome-test-builds/android/revisions/%s')).put()
    data_types.Job(
        name=BASIC_JOB_TYPE,
        environment_string=('HELP_URL = help_url\n')).put()
    data_types.Job(
        name=SRCMAP_JOB_TYPE,
        environment_string=(
            'PROJECT_NAME = libass\n'
            'HELP_URL = help_url\n'
            'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
            'blah-%s.srcmap.json')).put()
    data_types.Job(
        name=CUSTOM_BINARY_JOB_TYPE,
        environment_string=('CUSTOM_BINARY = True\n')).put()

  # General helper functions.
  @staticmethod
  def _read_data_file(data_file):
    """Helper function to read the contents of a data file."""
    with open(os.path.join(DATA_DIRECTORY, data_file)) as handle:
      return handle.read()

  # Helper classes and functions for mocks.
  class MockConfigChromium(object):
    """Simple mocked configuration for chromium."""

    def __init__(self):
      self.component_repository_mappings = 'default;chromium/src\nv8;v8/v8\n'

  class MockConfigOSSFuzz(object):
    """Simple mocked configuration."""

    def __init__(self):
      self.component_repository_mappings = ''

  @staticmethod
  def mock_get_url_content(url):
    """Read a local file based on the specified URL."""
    # Check to see if the URL specified is based on the default url.
    for line in REVISION_VARS_URL.splitlines():
      prefix = line[:line.find(';')]
      current_url = line[len(prefix) + 1:]
      format_index = current_url.find('%s')
      if url[:format_index] != current_url[:format_index]:
        continue

      revision_without_path = url[format_index:].split('/')[0]
      revision_without_path_and_query = revision_without_path.split('?')[0]
      deps_file = 'deps_%s_%s.txt' % (prefix, revision_without_path_and_query)
      data = RevisionsTestcase._read_data_file(deps_file)
      return data

    raise NotImplementedError(url)

  @staticmethod
  def mock_get_git_hash_for_git_commit_pos(git_commit_pos, _):
    """Return a fake git hash for a git commit position."""
    return hashlib.sha1(str(git_commit_pos).encode('utf-8')).hexdigest()

  # Tests.
  def test_convert_revision_to_integer_simple(self):
    """Test the simple revision case of convert_revision_to_integer."""
    revision = revisions.convert_revision_to_integer('12345')
    self.assertEqual(revision, 12345)

  @mock.patch('clusterfuzz._internal.metrics.logs.log_error')
  def test_convert_revision_to_integer_version_string(self, _):
    """Test version string conversions in convert_revision_to_integer."""
    revision = revisions.convert_revision_to_integer('1.1.1.1')

    # See the full comment in convert_revision_to_integer, but we pad this with
    # zeros to allow sorting.
    self.assertEqual(revision, 1000010000100001)

    # Ensure that the max lengths for each part are supported.
    revision = revisions.convert_revision_to_integer('12345.67890.12345.67890')
    self.assertEqual(revision, 12345678901234567890)

    # Ensure that we raise an exception if any of the individual parts are too
    # long.
    with self.assertRaises(ValueError):
      revisions.convert_revision_to_integer('123456.12345.12345.12345')

    with self.assertRaises(ValueError):
      revisions.convert_revision_to_integer('12345.123456.12345.12345')

    with self.assertRaises(ValueError):
      revisions.convert_revision_to_integer('12345.12345.123456.12345')

    with self.assertRaises(ValueError):
      revisions.convert_revision_to_integer('12345.12345.12345.123456')

    # Ensure that junk strings also raise value errors.
    with self.assertRaises(ValueError):
      revisions.convert_revision_to_integer('123junk')

    with self.assertRaises(ValueError):
      revisions.convert_revision_to_integer('junk')

    with self.assertRaises(ValueError):
      revisions.convert_revision_to_integer('junk123')

    with self.assertRaises(ValueError):
      revisions.convert_revision_to_integer('...')

  @mock.patch('clusterfuzz._internal.config.db_config.get')
  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  @mock.patch('clusterfuzz._internal.build_management.revisions.'
              '_git_commit_position_to_git_hash_for_chromium')
  def test_get_component_range_list_chromium(
      self, mock_get_git_hash, mock_get_url_content, mock_get_config):
    """Test that get_component_range_list works properly for the Chromium
    repo."""
    mock_get_config.return_value = self.MockConfigChromium()
    self.mock.default_project_name.return_value = 'chromium'
    mock_get_url_content.side_effect = self.mock_get_url_content
    mock_get_git_hash.side_effect = self.mock_get_git_hash_for_git_commit_pos

    result = revisions.get_component_range_list(336903, 336983, BASIC_JOB_TYPE)
    result_as_html = revisions.format_revision_list(result)

    expected_html = self._read_data_file('chromium_expected_html.txt')
    self.assertEqual(result_as_html, expected_html)

  @mock.patch('clusterfuzz._internal.config.db_config.get')
  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  @mock.patch('clusterfuzz._internal.build_management.revisions.'
              '_git_commit_position_to_git_hash_for_chromium')
  def test_get_component_range_list_clank(
      self, mock_get_git_hash, mock_get_url_content, mock_get_config):
    """Test that get_component_range_list works properly for the Clank repo."""
    mock_get_config.return_value = self.MockConfigChromium()
    self.mock.default_project_name.return_value = 'chromium'
    mock_get_url_content.side_effect = self.mock_get_url_content
    mock_get_git_hash.side_effect = self.mock_get_git_hash_for_git_commit_pos

    result = revisions.get_component_range_list(260548, 260552,
                                                ANDROID_JOB_TYPE)
    result_as_html = revisions.format_revision_list(result)

    expected_html = self._read_data_file('clank_expected_html.txt')
    self.assertEqual(result_as_html, expected_html)

  @mock.patch('clusterfuzz._internal.config.db_config.get')
  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  def test_get_git_hash_for_git_commit_pos(self, mock_get_url_content,
                                           mock_get_config):
    """Test git hash for git commit position."""
    mock_get_config.return_value = self.MockConfigChromium()
    self.mock.default_project_name.return_value = 'chromium'
    expected_hash = '95a3bc965ed80186215cea788caa5faae0898839'
    mock_get_url_content.return_value = ('{\n'
                                         '  "git_sha": "%s"\n'
                                         '}' % expected_hash)

    actual_hash = revisions._git_commit_position_to_git_hash_for_chromium(  # pylint: disable=protected-access
        '29124', 'v8/v8')
    self.assertEqual(actual_hash, expected_hash)

    # This is not perfect in that it assumes a specific order to the arguments,
    # but should be sufficient. A format string is not being used to make the
    # URL encoding easier to understand.
    mock_get_url_content.assert_called_once_with(
        revisions.CRREV_NUMBERING_URL +
        '?number=29124&numbering_identifier=refs%2Fheads%2Fmain&'
        'numbering_type=COMMIT_POSITION&project=chromium&'
        'repo=v8%2Fv8&fields=git_sha')

  @mock.patch('clusterfuzz._internal.config.db_config.get')
  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  @mock.patch('clusterfuzz._internal.build_management.revisions.'
              '_git_commit_position_to_git_hash_for_chromium')
  def test_get_real_revision_chromium(self, mock_get_git_hash,
                                      mock_get_url_content, mock_get_config):
    """Test that get_real_revision works properly for chromium revisions."""
    mock_get_config.return_value = self.MockConfigChromium()
    self.mock.default_project_name.return_value = 'chromium'
    mock_get_url_content.side_effect = self.mock_get_url_content
    mock_get_git_hash.side_effect = self.mock_get_git_hash_for_git_commit_pos

    self.assertEqual('1d783bc2a3629b94c963debfa3feaee27092dd92',
                     revisions.get_real_revision(336903, BASIC_JOB_TYPE))

  @mock.patch('clusterfuzz._internal.config.db_config.get')
  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  def test_get_real_revision_oss_fuzz(self, mock_get_url_content,
                                      mock_get_config):
    """Test that get_real_revision works properly for non-chromium revisions."""
    mock_get_config.return_value = self.MockConfigOSSFuzz()
    self.mock.default_project_name.return_value = 'oss-fuzz'
    mock_get_url_content.side_effect = self.mock_get_url_content

    self.assertEqual('35dc4dd0e14e3afb4a2c7e319a3f4110e20c7cf2',
                     revisions.get_real_revision(9002, SRCMAP_JOB_TYPE))
    self.assertEqual('54444451414e3efb4a4c7e319a3f4110e20c7cf2',
                     revisions.get_real_revision(9003, SRCMAP_JOB_TYPE))

  @mock.patch('clusterfuzz._internal.config.db_config.get')
  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  def test_get_component_revision_list_src_map(self, mock_get_url_content,
                                               mock_get_config):
    """Test get_component_range_list for srcmap jobs."""
    mock_get_config.return_value = self.MockConfigOSSFuzz()
    self.mock.default_project_name.return_value = 'oss-fuzz'
    mock_get_url_content.side_effect = self.mock_get_url_content

    result = revisions.get_component_range_list(1337, 9001, SRCMAP_JOB_TYPE)
    result_as_html = revisions.format_revision_list(result)
    expected_html = self._read_data_file('srcmap_expected_html.txt')
    self.assertEqual(result_as_html, expected_html)

  @mock.patch('clusterfuzz._internal.config.db_config.get')
  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  def test_get_component_revision_list_src_map_text(self, mock_get_url_content,
                                                    mock_get_config):
    """Test get_component_range_list for srcmap jobs (text only)."""
    mock_get_config.return_value = self.MockConfigOSSFuzz()
    self.mock.default_project_name.return_value = 'oss-fuzz'
    mock_get_url_content.side_effect = self.mock_get_url_content

    result = revisions.get_component_range_list(1337, 9001, SRCMAP_JOB_TYPE)
    result_as_html = revisions.format_revision_list(result, use_html=False)
    expected_html = self._read_data_file('srcmap_expected_text.txt')
    self.assertEqual(result_as_html, expected_html)

  @mock.patch('clusterfuzz._internal.config.db_config.get')
  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  def test_get_component_range_list_same_hash(self, mock_get_url_content,
                                              mock_get_config):
    """Test get_component_range_list for 2 builds that have different revision
    numbers, but same revision hash after mapping."""
    mock_get_config.return_value = self.MockConfigOSSFuzz()
    self.mock.default_project_name.return_value = 'oss-fuzz'
    mock_get_url_content.side_effect = self.mock_get_url_content

    result = revisions.get_component_range_list(1337, 1338, SRCMAP_JOB_TYPE)
    result_as_html = revisions.format_revision_list(result)
    expected_html = self._read_data_file('srcmap_expected_html_2.txt')
    self.assertEqual(result_as_html, expected_html)

  @mock.patch('clusterfuzz._internal.config.db_config.get')
  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  def test_get_component_range_list_0_start_custom(self, mock_get_url_content,
                                                   mock_get_config):
    """Test get_component_range_list with a '0' start_revision."""
    mock_get_config.return_value = self.MockConfigOSSFuzz()
    self.mock.default_project_name.return_value = 'oss-fuzz'
    mock_get_url_content.side_effect = self.mock_get_url_content

    result = revisions.get_component_range_list(0, 1338, SRCMAP_JOB_TYPE)
    result_as_html = revisions.format_revision_list(result)
    expected_html = self._read_data_file('srcmap_expected_html_3.txt')
    self.assertEqual(result_as_html, expected_html)

  @mock.patch('clusterfuzz._internal.config.db_config.get')
  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  def test_get_component_range_list_custom_binary_job(
      self, mock_get_url_content, mock_get_config):
    """Test get_component_range_list with a custom binary job."""
    mock_get_config.return_value = self.MockConfigChromium()
    self.mock.default_project_name.return_value = 'chromium'
    mock_get_url_content.side_effect = self.mock_get_url_content

    self.assertEqual([],
                     revisions.get_component_range_list(1337, 1338,
                                                        CUSTOM_BINARY_JOB_TYPE))

  def test_find_build_url(self):
    """Tests test_find_build_url."""
    bucket_path = 'gs://bucket/path/asan-([0-9]+).zip'
    build_urls_list = [
        'gs://bucket/path/asan-1000.zip',
        'gs://bucket/path/asan-2000.zip',
        'gs://bucket/path/asan-3000.zip',
        'gs://bucket/path/asan-4000.zip',
    ]
    self.assertEqual(
        revisions.find_build_url(bucket_path, build_urls_list, 1000),
        'gs://bucket/path/asan-1000.zip')
    self.assertEqual(
        revisions.find_build_url(bucket_path, build_urls_list, 2000),
        'gs://bucket/path/asan-2000.zip')
    self.assertEqual(
        revisions.find_build_url(bucket_path, build_urls_list, 3000),
        'gs://bucket/path/asan-3000.zip')
    self.assertEqual(
        revisions.find_build_url(bucket_path, build_urls_list, 4000),
        'gs://bucket/path/asan-4000.zip')
    self.assertIsNone(
        revisions.find_build_url(bucket_path, build_urls_list, 5000))

  def test_find_min_revision_index(self):
    """Tests find_min_revision_index()."""
    revisions_list = [1000, 2000, 3000, 4000]
    self.assertEqual(revisions.find_min_revision_index(revisions_list, 1000), 0)
    self.assertEqual(revisions.find_min_revision_index(revisions_list, 2000), 1)
    self.assertEqual(revisions.find_min_revision_index(revisions_list, 3000), 2)
    self.assertEqual(revisions.find_min_revision_index(revisions_list, 4000), 3)

    self.assertIsNone(revisions.find_min_revision_index(revisions_list, 1))
    self.assertEqual(revisions.find_min_revision_index(revisions_list, 1001), 0)
    self.assertEqual(revisions.find_min_revision_index(revisions_list, 2001), 1)
    self.assertEqual(revisions.find_min_revision_index(revisions_list, 3001), 2)
    self.assertEqual(revisions.find_min_revision_index(revisions_list, 4001), 3)

  def test_find_max_revision_index(self):
    """Tests find_max_revision_index()."""
    revisions_list = [1000, 2000, 3000, 4000]
    self.assertEqual(revisions.find_max_revision_index(revisions_list, 1000), 0)
    self.assertEqual(revisions.find_max_revision_index(revisions_list, 2000), 1)
    self.assertEqual(revisions.find_max_revision_index(revisions_list, 3000), 2)
    self.assertEqual(revisions.find_max_revision_index(revisions_list, 4000), 3)

    self.assertEqual(revisions.find_max_revision_index(revisions_list, 1), 0)
    self.assertEqual(revisions.find_max_revision_index(revisions_list, 1001), 1)
    self.assertEqual(revisions.find_max_revision_index(revisions_list, 2001), 2)
    self.assertEqual(revisions.find_max_revision_index(revisions_list, 3001), 3)
    self.assertIsNone(revisions.find_max_revision_index(revisions_list, 4001))

  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  def test_revision_to_branched_from(self, mock_get_url_content):
    """Test revision_to_branched_from() with normal branch."""
    mock_get_url_content.side_effect = self.mock_get_url_content
    val = revisions.revision_to_branched_from('https://test-repository/src.git',
                                              'abcd')
    self.assertEqual('36884', val)

  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  def test_revision_to_branched_from_no_branched_from(self,
                                                      mock_get_url_content):
    """Test revision_to_branched_from() where commit has no branch-from"""
    mock_get_url_content.side_effect = self.mock_get_url_content
    val = revisions.revision_to_branched_from('https://test-repository/src.git',
                                              'f00d')
    self.assertEqual(None, val)

  @mock.patch(
      'clusterfuzz._internal.build_management.revisions._get_url_content')
  def test_revision_to_branched_from_no_message(self, mock_get_url_content):
    """Test revision_to_branched_from() with no commit message at all"""
    mock_get_url_content.side_effect = self.mock_get_url_content
    val = revisions.revision_to_branched_from('https://test-repository/src.git',
                                              'd00d')
    self.assertEqual(None, val)


@test_utils.with_cloud_emulators('datastore')
class GetComponentsListTest(unittest.TestCase):
  """Tests get_components_list."""

  def setUp(self):
    helpers.patch_environ(self)

  def test_get_components_list(self):
    """Test get_components_list."""
    data_types.Job(
        name='libfuzzer_asan_libass',
        environment_string=('PROJECT_NAME = libass\n'
                            'HELP_URL = help_url\n')).put()
    revisions_dict = {
        u'/src/libass': {
            u'url': u'https://github.com/libass/libass.git',
            u'rev': u'35dc4dd0e14e3afb4a2c7e319a3f4110e20c7cf2',
        },
        u'/src/fribidi': {
            u'url': u'https://github.com/behdad/fribidi.git',
            u'rev': u'881b8d891cc61989ab8811b74d0e721f72bf913b',
        }
    }

    expected_components_list = [u'/src/libass', u'/src/fribidi']
    actual_components_list = revisions.get_components_list(
        revisions_dict, 'libfuzzer_asan_libass')
    self.assertEqual(expected_components_list, actual_components_list)

  def test_get_components_list_main_repo(self):
    """Test get_components_list with a main_repo set."""
    data_types.Job(
        name='libfuzzer_asan_project',
        environment_string=('PROJECT_NAME = project\n'
                            'MAIN_REPO = https://github.com/org/main.git\n'
                            'HELP_URL = help_url\n')).put()
    revisions_dict = {
        '/src/main': {
            'url': 'https://github.com/org/main.git',
            'rev': '35dc4dd0e14e3afb4a2c7e319a3f4110e20c7cf2',
        },
        '/src/project-fuzzing-corpus': {
            'url': 'https://github.com/org/project-fuzzing-corpus.git',
            'rev': '881b8d891cc61989ab8811b74d0e721f72bf913b',
        }
    }

    expected_components_list = ['/src/main', '/src/project-fuzzing-corpus']
    actual_components_list = revisions.get_components_list(
        revisions_dict, 'libfuzzer_asan_project')
    self.assertEqual(expected_components_list, actual_components_list)

  def test_get_components_list_no_exact_matches(self):
    """Test get_components_list (no exact matches)."""
    data_types.Job(
        name='libfuzzer_asan_php',
        environment_string=('PROJECT_NAME = php\n'
                            'HELP_URL = help_url\n')).put()
    revisions_dict = {
        '/src/afl': {
            'type': 'git',
            'url': 'https://github.com/AFLplusplus/AFLplusplus.git',
            'rev': '7cfa690d1c677e04e0935d6ac3c6b94b5fed50ad'
        },
        '/src/PHP-src': {
            'type': 'git',
            'url': 'https://github.com/php/php-src.git',
            'rev': '853b7945bc6c97d7d1643f5f8b22851e323829cd'
        },
        '/src/php-src/oniguruma': {
            'type': 'git',
            'url': 'https://github.com/kkos/oniguruma.git',
            'rev': '2b7b94122c696ffb5ed7fbe9a42c73d4563e3498'
        },
    }

    expected_components_list = [
        '/src/PHP-src', '/src/afl', '/src/php-src/oniguruma'
    ]
    actual_components_list = revisions.get_components_list(
        revisions_dict, 'libfuzzer_asan_php')
    self.assertEqual(expected_components_list, actual_components_list)

    revisions_dict = {
        '/src/afl': {
            'type': 'git',
            'url': 'https://github.com/AFLplusplus/AFLplusplus.git',
            'rev': '7cfa690d1c677e04e0935d6ac3c6b94b5fed50ad'
        },
        '/src/php-src': {
            'type': 'git',
            'url': 'https://github.com/php/php-src.git',
            'rev': '853b7945bc6c97d7d1643f5f8b22851e323829cd'
        },
        '/src/php': {
            'type': 'git',
            'url': 'https://github.com/kkos/oniguruma.git',
            'rev': '2b7b94122c696ffb5ed7fbe9a42c73d4563e3498'
        },
    }

    expected_components_list = ['/src/php', '/src/afl', '/src/php-src']
    actual_components_list = revisions.get_components_list(
        revisions_dict, 'libfuzzer_asan_php')
    self.assertEqual(expected_components_list, actual_components_list)


class DepsToRevisionsDictTest(unittest.TestCase):
  """Tests deps_to_revisions_dict"""

  # General helper function.
  @staticmethod
  def _read_data_file(data_file):
    """Helper function to read the contents of a data file."""
    with open(os.path.join(DATA_DIRECTORY, data_file)) as handle:
      return handle.read()

  def test(self):
    """Test that deps is correctly parsed without exceptions."""
    self.maxDiff = None  # pylint: disable=invalid-name
    deps_content = self._read_data_file('chromium_deps.txt')
    actual_revisions_dict = revisions.deps_to_revisions_dict(deps_content)
    expected_revisions_dict = ast.literal_eval(
        self._read_data_file('chromium_expected_deps_revisions_dict.txt'))
    self.assertEqual(expected_revisions_dict, actual_revisions_dict)

  def test_bad_deps(self):
    """Test that bad deps is correctly parsed without exceptions."""
    self.maxDiff = None  # pylint: disable=invalid-name
    deps_content = 'vars = {}'
    actual_revisions_dict = revisions.deps_to_revisions_dict(deps_content)
    self.assertEqual(None, actual_revisions_dict)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Write arguments for launcher.py to flags file.
"""

import os

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import builtin
from clusterfuzz._internal.system import environment

# The name of the file that run.py puts in the input directory so AFL doesn't
# complain, since it needs at least one input file.
AFL_DUMMY_INPUT = 'in1'


def write_dummy_file(input_dir):
  """Afl will refuse to run if the corpus directory is empty or contains empty
  files. So write the bare minimum to get afl to run if there is no corpus
  yet."""
  # TODO(metzman): Ask lcamtuf to allow AFL to run with an empty input corpus.
  dummy_input_path = os.path.join(input_dir, AFL_DUMMY_INPUT)
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    file_host.write_data_to_worker(b' ', dummy_input_path)
  else:
    utils.write_data_to_file(' ', dummy_input_path)


class Afl(builtin.EngineFuzzer):
  """Builtin AFL fuzzer."""

  def generate_arguments(self, fuzzer_path):  # pylint: disable=unused-argument
    """Generate arguments for fuzzer using .options file or default values."""
    return ''

  def run(self, input_directory, output_directory, no_of_files):
    result = super(Afl, self).run(input_directory, output_directory,
                                  no_of_files)

    write_dummy_file(result.corpus_directory)
    return result
